"""Tests for applying masks to swiftgalaxy, datasets and named columns."""

import pytest
from copy import copy, deepcopy
import numpy as np
from unyt.testing import assert_allclose_units
from swiftgalaxy import SWIFTGalaxy, MaskCollection
from swiftgalaxy.demo_data import (
    ToyHF,
    _create_toysnap,
    _remove_toysnap,
    _toysnap_filename,
    _present_particle_types,
    _n_g_all,
    _n_dm_all,
    _n_s_all,
    _n_bh_all,
)
from swiftgalaxy.masks import LazyMask

abstol_nd = 1.0e-4
reltol_nd = 1.0e-4


class TestMaskingSWIFTGalaxy:
    """Test applying masks to :class:`~swiftgalaxy.reader.SWIFTGalaxy` objects."""

    @pytest.mark.parametrize("particle_name", _present_particle_types.values())
    def test_getattr_masking(self, sg, particle_name):
        """Test masking with square bracket notation. Uses an order reversing mask."""
        getattr(sg, particle_name).particle_ids
        mask = np.s_[::-1]
        new_sg = sg[MaskCollection(**{particle_name: mask})]
        assert (
            getattr(new_sg, particle_name).particle_ids[::-1]
            == getattr(sg, particle_name).particle_ids
        ).all()

    @pytest.mark.parametrize("particle_name", _present_particle_types.values())
    @pytest.mark.parametrize("before_load", (True, False))
    def test_reordering_slice_mask(self, sg, particle_name, before_load):
        """Test whether a slice mask that re-orders elements works."""
        mask = np.s_[::-1]
        ids_before = getattr(sg, particle_name).particle_ids
        if before_load:
            getattr(sg, particle_name)._particle_dataset._particle_ids = None
            del getattr(sg._extra_mask, particle_name)._mask
            getattr(sg._extra_mask, particle_name)._evaluated = False
        sg.mask_particles(MaskCollection(**{particle_name: mask}))
        ids = getattr(sg, particle_name).particle_ids
        assert_allclose_units(ids_before[mask], ids, rtol=0, atol=0)

    @pytest.mark.parametrize("particle_name", _present_particle_types.values())
    @pytest.mark.parametrize("before_load", (True, False))
    def test_reordering_int_mask(self, sg, particle_name, before_load):
        """Test an integer array mask that re-orders elements and changes the length."""
        ids_before = getattr(sg, particle_name).particle_ids
        mask = np.arange(ids_before.size)
        # randomize order (in-place operation)
        np.random.shuffle(mask)
        # keep half the particles
        mask = mask[: mask.size // 2]
        if before_load:
            getattr(sg, particle_name)._particle_dataset._particle_ids = None
            del getattr(sg._extra_mask, particle_name)._mask
            getattr(sg._extra_mask, particle_name)._evaluated = False
        sg.mask_particles(MaskCollection(**{particle_name: mask}))
        ids = getattr(sg, particle_name).particle_ids
        assert_allclose_units(ids_before[mask], ids, rtol=0, atol=0)

    @pytest.mark.parametrize("particle_name", _present_particle_types.values())
    @pytest.mark.parametrize("before_load", (True, False))
    def test_bool_mask(self, sg, particle_name, before_load):
        """Test whether a boolean array mask works."""
        ids_before = getattr(sg, particle_name).particle_ids
        # randomly keep about half of particles
        mask = np.random.rand(ids_before.size) > 0.5
        if before_load:
            getattr(sg, particle_name)._particle_dataset._particle_ids = None
            del getattr(sg._extra_mask, particle_name)._mask
            getattr(sg._extra_mask, particle_name)._evaluated = False
        sg.mask_particles(MaskCollection(**{particle_name: mask}))
        ids = getattr(sg, particle_name).particle_ids
        assert_allclose_units(ids_before[mask], ids, rtol=0, atol=0)

    @pytest.mark.parametrize("before_load", (True, False))
    def test_namedcolumn_masked(self, sg, before_load):
        """Test that named columns get masked too."""
        neutral_before = sg.gas.hydrogen_ionization_fractions.neutral
        mask = np.random.rand(neutral_before.size) > 0.5
        if before_load:
            sg.gas.hydrogen_ionization_fractions._named_column_dataset._neutral = None
            del sg._extra_mask.gas._mask
            sg._extra_mask.gas._evaluated = False
        sg.mask_particles(MaskCollection(**{"gas": mask}))
        neutral = sg.gas.hydrogen_ionization_fractions.neutral
        assert_allclose_units(
            neutral_before[mask], neutral, rtol=reltol_nd, atol=abstol_nd
        )

    def test_mask_without_spatial_mask(self, tmp_path_factory):
        """
        Check that if we have no masks we read everything in the box (and warn about it).

        Then that we can still apply an extra mask, and a second one (there's specific
        logic for applying two consecutively).
        """
        toysnap_filename = (
            tmp_path_factory.mktemp(_toysnap_filename.parent) / _toysnap_filename.name
        )
        try:
            _create_toysnap(snapfile=toysnap_filename)
            sg = SWIFTGalaxy(
                toysnap_filename,
                None,  # no halo_catalogue is easiest way to get no mask
                transforms_like_coordinates={"coordinates", "extra_coordinates"},
                transforms_like_velocities={"velocities", "extra_velocities"},
            )
            # check that extra mask is blank for all particle types:
            assert sg._extra_mask.gas is None
            assert sg._extra_mask.dark_matter is None
            assert sg._extra_mask.stars is None
            assert sg._extra_mask.black_holes is None
            # check that cell mask is blank for all particle types:
            assert sg._spatial_mask is None
            # check that we read all the particles:
            assert sg.gas.masses.size == _n_g_all
            assert sg.dark_matter.masses.size == _n_dm_all
            assert sg.stars.masses.size == _n_s_all
            assert sg.black_holes.masses.size == _n_bh_all
            # now apply an extra mask
            sg.mask_particles(MaskCollection(gas=np.s_[:1000]))
            assert sg.gas.masses.size == 1000
            # and the second consecutive one
            sg.mask_particles(MaskCollection(gas=np.s_[:100]))
            assert sg.gas.masses.size == 100
        finally:
            _remove_toysnap(snapfile=toysnap_filename)


class TestMaskingParticleDatasets:
    """Test applying masks to particle datasets."""

    @pytest.mark.parametrize("particle_name", _present_particle_types.values())
    @pytest.mark.parametrize("before_load", (True, False))
    def test_reordering_slice_mask(self, sg, particle_name, before_load):
        """Test whether a slice mask that re-orders elements works."""
        mask = np.s_[::-1]
        ids_before = getattr(sg, particle_name).particle_ids
        if before_load:
            getattr(sg, particle_name)._particle_dataset._particle_ids = None
            del getattr(sg._extra_mask, particle_name)._mask
            getattr(sg._extra_mask, particle_name)._evaluated = False
        masked_dataset = getattr(sg, particle_name)[mask]
        ids = masked_dataset.particle_ids
        assert_allclose_units(ids_before[mask], ids, rtol=0, atol=0)

    @pytest.mark.parametrize("particle_name", _present_particle_types.values())
    @pytest.mark.parametrize("before_load", (True, False))
    def test_reordering_int_mask(self, sg, particle_name, before_load):
        """Test masking with an integer array: re-orders elements and changes length."""
        ids_before = getattr(sg, particle_name).particle_ids
        mask = np.arange(ids_before.size)
        # randomize order (in-place operation)
        np.random.shuffle(mask)
        # keep half the particles
        mask = mask[: mask.size // 2]
        if before_load:
            getattr(sg, particle_name)._particle_dataset._particle_ids = None
            del getattr(sg._extra_mask, particle_name)._mask
            getattr(sg._extra_mask, particle_name)._evaluated = False
        masked_dataset = getattr(sg, particle_name)[mask]
        ids = masked_dataset.particle_ids
        assert_allclose_units(ids_before[mask], ids, rtol=0, atol=0)

    @pytest.mark.parametrize("particle_name", _present_particle_types.values())
    @pytest.mark.parametrize("before_load", (True, False))
    def test_bool_mask(self, sg, particle_name, before_load):
        """Test whether a boolean array mask works."""
        ids_before = getattr(sg, particle_name).particle_ids
        # randomly keep about half of particles
        mask = np.random.rand(ids_before.size) > 0.5
        if before_load:
            getattr(sg, particle_name)._particle_dataset._particle_ids = None
            del getattr(sg._extra_mask, particle_name)._mask
            getattr(sg._extra_mask, particle_name)._evaluated = False
        masked_dataset = getattr(sg, particle_name)[mask]
        ids = masked_dataset.particle_ids
        assert_allclose_units(ids_before[mask], ids, rtol=0, atol=0)


class TestMaskingNamedColumnDatasets:
    """Test applying masks to named column datasets."""

    @pytest.mark.parametrize("before_load", (True, False))
    def test_reordering_slice_mask(self, sg, before_load):
        """Test whether a slice mask that re-orders elements works."""
        mask = np.s_[::-1]
        fractions_before = sg.gas.hydrogen_ionization_fractions.neutral
        if before_load:
            sg.gas.hydrogen_ionization_fractions._neutral = None
            del sg._extra_mask.gas._mask
            sg._extra_mask.gas._evaluated = False
        masked_namedcolumnsdataset = sg.gas.hydrogen_ionization_fractions[mask]
        fractions = masked_namedcolumnsdataset.neutral
        assert_allclose_units(
            fractions_before[mask], fractions, rtol=reltol_nd, atol=abstol_nd
        )

    @pytest.mark.parametrize("before_load", (True, False))
    def test_reordering_int_mask(self, sg, before_load):
        """Test masking with an integer array: re-orders and changes the length."""
        fractions_before = sg.gas.hydrogen_ionization_fractions.neutral
        mask = np.arange(fractions_before.size)
        # randomize order (in-place operation)
        np.random.shuffle(mask)
        # keep half the particles
        mask = mask[: mask.size // 2]
        if before_load:
            sg.gas.hydrogen_ionization_fractions._neutral = None
            del sg._extra_mask.gas._mask
            sg._extra_mask.gas._mask = False
        masked_namedcolumnsdataset = sg.gas.hydrogen_ionization_fractions[mask]
        fractions = masked_namedcolumnsdataset.neutral
        assert_allclose_units(
            fractions_before[mask], fractions, rtol=reltol_nd, atol=abstol_nd
        )

    @pytest.mark.parametrize("before_load", (True, False))
    def test_bool_mask(self, sg, before_load):
        """Test whether a boolean array mask works."""
        fractions_before = sg.gas.hydrogen_ionization_fractions.neutral
        # randomly keep about half of particles
        mask = np.random.rand(fractions_before.size) > 0.5
        if before_load:
            sg.gas.hydrogen_ionization_fractions._neutral = None
            del sg._extra_mask.gas._mask
            sg._extra_mask.gas._evaluated = False
        masked_namedcolumnsdataset = sg.gas.hydrogen_ionization_fractions[mask]
        fractions = masked_namedcolumnsdataset.neutral
        assert_allclose_units(
            fractions_before[mask], fractions, rtol=reltol_nd, atol=abstol_nd
        )


class TestMultiModeMask:
    """Tests handling a halo catalogue with multiple galaxies selected."""

    def test_mask_multi_invalid(self, sg):
        """
        Check that catalogues raise when accessing single row in multi mode.

        When a multi-galaxy halo catalogue object is not masked down to a single object,
        attempting to request a mask for a single object should raise.
        """
        cat = ToyHF(index=[0, 1])
        assert cat._multi_galaxy  # confirm this is multi-galaxy mode
        with pytest.raises(
            RuntimeError,
            match="Halo catalogue has multiple galaxies and is not currently masked.",
        ):
            cat._get_extra_mask(sg)

    def test_caesar_mask_catalogue(self, caesar_multi):
        """
        Check that caesar catalogues raise when accessing single row in multi mode.

        If a caesar catalogue is in multi-galaxy mode and is not currently masked then
        trying to use the helper to select a single row when the catalogue isn't currently
        restricted to a single galaxy should raise.
        """
        with pytest.raises(
            RuntimeError, match="Tried to mask catalogue without mask index!"
        ):
            caesar_multi._mask_catalogue()


class TestLazyMask:
    """Unit tests for the LazyMask class itself."""

    def test_init_lazy(self, lm):
        """Check that initializing in lazy mode works."""
        assert lm._evaluated is False
        assert lm._mask_function is not None  # also implies it exists
        assert not hasattr(lm, "_mask")

    def test_init_concrete(self):
        """Check that initializing with a concrete mask works."""
        m = np.ones(10, dtype=bool)
        lm = LazyMask(mask=m)
        assert lm._evaluated
        assert lm._mask is m
        assert lm._mask_function is None

    def test_init_none(self):
        """Check that initializing with ``None`` works (i.e. no mask)."""
        lm = LazyMask(mask=None)
        assert lm._evaluated
        assert lm._mask is None
        assert not hasattr(lm, "_mask_function")

    def test_trigger_eval(self, lm):
        """Check that accessing mask triggers evaluation if lazy."""
        assert lm._evaluated is False
        assert not hasattr(lm, "_mask")
        assert (lm.mask == lm._mask_function()).all()
        assert lm._evaluated
        assert (lm._mask == lm._mask_function()).all()

    def test_access_not_lazy(self):
        """Check that accessing the mask works for a non-lazy mask."""
        m = np.ones(10, dtype=bool)
        lm = LazyMask(mask=m)
        assert lm._evaluated
        assert (lm.mask == m).all()

    def test_manual_trigger_eval(self, lm):
        """Check that accessing mask triggers evaluation if lazy."""
        assert lm._evaluated is False
        assert not hasattr(lm, "_mask")
        lm._evaluate()
        assert lm._evaluated
        assert (lm.mask == lm._mask_function()).all()
        assert (lm._mask == lm._mask_function()).all()

    def test_trigger_eval_once_only(self):
        """Check that we can't trigger mask evaluation repeatedly."""

        class MF(object):
            """A mask generator that counts how many times its ``__call__`` is called."""

            call_counter: int = 0

            def __call__(self):
                """
                Call the class to behave like a simple mask function.

                Returns
                -------
                out : ndarray
                    A simple mask array.
                """
                self.call_counter += 1
                return np.ones(10, dtype=bool)

        mf = MF()
        lm = LazyMask(mask_function=mf)
        assert lm._evaluated is False
        # trigger a mask evaluation:
        lm.mask
        assert lm._evaluated is True
        assert mf.call_counter == 1
        # we shouldn't be able to trigger or force another mask evaluation:
        lm.mask
        lm._evaluate()
        assert mf.call_counter == 1

    def test_copy(self, lm):
        """Check copying behaviour of ``LazyMask`` objects."""
        # first copy before evaluating
        lm_unevaluated_copy = copy(lm)
        assert lm_unevaluated_copy._evaluated is False
        assert not hasattr(lm_unevaluated_copy, "_mask")
        assert lm_unevaluated_copy._mask_function is lm._mask_function
        # trigger evaluated
        lm._evaluate()
        assert lm._evaluated
        # now copy after evaluating
        lm_evaluated_copy = copy(lm)
        assert lm_evaluated_copy._evaluated
        assert (lm_evaluated_copy._mask == lm._mask_function()).all()
        assert lm_evaluated_copy._mask_function is lm._mask_function
        # and test a non-lazy mask
        m = np.ones(10, dtype=bool)
        nlm = LazyMask(mask=m)
        nlm_copy = copy(nlm)
        assert nlm_copy._evaluated
        assert (nlm_copy._mask == m).all()
        assert nlm_copy._mask_function is None

    def test_deepcopy(self, lm):
        """Check deep copying behaviour of ``LazyMask`` objects."""
        # first copy before evaluating
        lm_unevaluated_copy = deepcopy(lm)
        assert lm_unevaluated_copy._evaluated is False
        assert not hasattr(lm_unevaluated_copy, "_mask")
        assert lm_unevaluated_copy._mask_function is lm._mask_function
        # trigger evaluated
        lm._evaluate()
        assert lm._evaluated
        # now copy after evaluating
        lm_evaluated_copy = deepcopy(lm)
        assert lm_evaluated_copy._evaluated
        assert (lm_evaluated_copy._mask == lm._mask_function()).all()
        assert lm_evaluated_copy._mask_function is lm._mask_function
        # and test a non-lazy mask
        m = np.ones(10, dtype=bool)
        nlm = LazyMask(mask=m)
        nlm_copy = copy(nlm)
        assert nlm_copy._evaluated
        assert (nlm_copy._mask == m).all()
        assert nlm_copy._mask_function is None

    def test_compare_lazymask(self, lm):
        """Check comparison behaviour between two ``LazyMask`` objects."""
        lm2 = copy(lm)
        with pytest.raises(
            ValueError, match="Cannot compare when one or more masks are not evaluated."
        ):
            lm == lm2
        lm._evaluate()
        lm2._evaluate()
        assert lm == lm2
        lmn = LazyMask(mask=None)
        assert lm != lmn

    def test_compare_nonlazymask(self, lm):
        """Check comparison behaviour between a ``LazyMask`` and other objects."""
        m = np.ones(10, dtype=bool)
        with pytest.raises(
            ValueError, match="Cannot compare when one or more masks are not evaluated."
        ):
            lm == m
        lm._evaluate()
        assert lm == m
        assert lm != np.zeros(10, dtype=bool)

    def test_compare_nonemask(self):
        """Check comparison behaviour between null masks."""
        lm = LazyMask(mask=None)
        assert lm == lm
        assert not lm != lm



"""Tests of integration of swiftsimio visualisation tools with swiftgalaxy."""

import pytest
import numpy as np
import unyt as u
from swiftsimio import cosmo_array, cosmo_quantity
from swiftsimio.visualisation.projection import project_gas
from swiftsimio.visualisation.slice import slice_gas
from swiftsimio.visualisation.volume_render import render_gas
from swiftgalaxy import SWIFTGalaxy


@pytest.mark.parametrize("periodic", [False, True])
class TestRecenteredVisualisation:
    """Tests for visualisation functions on swiftgalaxies with recentered coordinates."""

    @pytest.mark.parametrize("z_cut", [False, True])
    def test_recentered_projection(self, sg_autorecentre_off, z_cut, periodic):
        """We should be able to make the same projections whether we recentered or not."""
        sg = SWIFTGalaxy(
            sg_autorecentre_off.snapshot_filename, sg_autorecentre_off.halo_catalogue
        )
        z_limits = [1.9, 2.1] if z_cut else []
        recentered_z_limits = [-0.1, 0.1] if z_cut else []
        kwargs = {
            "resolution": 4,
            "project": "masses",
            "parallel": True,
            "periodic": periodic,
        }
        ref_img = project_gas(
            sg_autorecentre_off,
            region=cosmo_array(
                [1.5, 2.5, 1.5, 2.5] + z_limits,
                u.Mpc,
                comoving=True,
                scale_factor=1,
                scale_exponent=1,
            ),
            **kwargs,
        )
        recentered_img = project_gas(
            sg,
            region=cosmo_array(
                [-0.5, 0.5, -0.5, 0.5] + recentered_z_limits,
                u.Mpc,
                comoving=True,
                scale_factor=1,
                scale_exponent=1,
            ),
            **kwargs,
        )
        assert (ref_img > 0).any()
        assert np.allclose(ref_img, recentered_img)

    def test_recentered_slice(self, sg_autorecentre_off, periodic):
        """We should be able to make the same slice whether we recentered or not."""
        sg = SWIFTGalaxy(
            sg_autorecentre_off.snapshot_filename, sg_autorecentre_off.halo_catalogue
        )
        z_slice = cosmo_quantity(
            1.95, u.Mpc, comoving=True, scale_factor=1, scale_exponent=1
        )
        recentered_z_slice = cosmo_quantity(
            -0.05, u.Mpc, comoving=True, scale_factor=1, scale_exponent=1
        )
        kwargs = {
            "resolution": 4,
            "project": "masses",
            "parallel": True,
            "periodic": periodic,
        }
        ref_img = slice_gas(
            sg_autorecentre_off,
            z_slice=z_slice,
            region=cosmo_array(
                [1.5, 2.5, 1.5, 2.5],
                u.Mpc,
                comoving=True,
                scale_factor=1,
                scale_exponent=1,
            ),
            **kwargs,
        )
        recentered_img = slice_gas(
            sg,
            z_slice=recentered_z_slice,
            region=cosmo_array(
                [-0.5, 0.5, -0.5, 0.5],
                u.Mpc,
                comoving=True,
                scale_factor=1,
                scale_exponent=1,
            ),
            **kwargs,
        )
        assert (ref_img > 0).any()
        assert np.allclose(ref_img, recentered_img)

    def test_recentered_volume_render(self, sg_autorecentre_off, periodic):
        """We should be able to make the same rendering whether we recentered or not."""
        sg = SWIFTGalaxy(
            sg_autorecentre_off.snapshot_filename, sg_autorecentre_off.halo_catalogue
        )
        kwargs = {
            "resolution": 4,
            "project": "masses",
            "parallel": True,
            "periodic": periodic,
        }
        ref_img = render_gas(
            sg_autorecentre_off,
            region=cosmo_array(
                [1.5, 2.5, 1.5, 2.5, 1.5, 2.5],
                u.Mpc,
                comoving=True,
                scale_factor=1,
                scale_exponent=1,
            ),
            **kwargs,
        )
        recentered_img = render_gas(
            sg,
            region=cosmo_array(
                [-0.5, 0.5, -0.5, 0.5, -0.5, 0.5],
                u.Mpc,
                comoving=True,
                scale_factor=1,
                scale_exponent=1,
            ),
            **kwargs,
        )
        assert (ref_img > 0).any()
        assert np.allclose(ref_img, recentered_img)



"""Test running the examples that come with :mod:`swiftgalaxy`."""

from pathlib import Path
import pytest
import unyt as u
from swiftsimio import cosmo_array, mask
from swiftgalaxy import SWIFTGalaxy, Velociraptor, Caesar, SOAP, Standalone
from swiftgalaxy.demo_data import (
    ToyHF,
    _toysnap_filename,
    _toyvr_filebase,
    _toysoap_filename,
    _toysoap_membership_filebase,
    _toysoap_virtual_snapshot_filename,
    _toycaesar_filename,
)


class TestWebExampleData:
    """Test the example data available for download."""

    def test_str(self, web_examples_tmpdir):
        """Check that string representation lists available examples."""
        assert "snapshot" in str(web_examples_tmpdir)

    def test_snapshot(self, web_examples_tmpdir):
        """Check that we can create a swiftgalaxy, retrieving a sample snapshot file."""
        sm = mask(web_examples_tmpdir.snapshot)  # just to get an interface to metadata
        SWIFTGalaxy(
            web_examples_tmpdir.snapshot,
            Standalone(
                centre=cosmo_array(
                    [2, 2, 2],
                    u.Mpc,
                    comoving=True,
                    scale_factor=sm.metadata.scale_factor,
                    scale_exponent=1,
                ),
                velocity_centre=cosmo_array(
                    [200, 200, 200],
                    u.km / u.s,
                    comoving=True,
                    scale_factor=sm.metadata.scale_factor,
                    scale_exponent=0,
                ),
                spatial_offsets=cosmo_array(
                    [[-1, 1], [-1, 1], [-1, 1]],
                    u.Mpc,
                    comoving=True,
                    scale_factor=sm.metadata.scale_factor,
                    scale_exponent=1,
                ),
            ),
        )

    @pytest.mark.parametrize("group_type", ["halo", "galaxy"])
    def test_caesar(self, web_examples_tmpdir, group_type):
        """
        Check that we can create a swiftgalaxy.

        We should retrieve a sample snapshot file and caesar catalogue.
        """
        pytest.importorskip("caesar")
        SWIFTGalaxy(
            web_examples_tmpdir.snapshot,
            Caesar(web_examples_tmpdir.caesar, group_type=group_type, group_index=0),
        )

    def test_soap(self, web_examples_tmpdir):
        """
        Check that we can create a swiftgalaxy.

        We should retrieve a sample virtual snapshot file and a soap catalogue.
        """
        SWIFTGalaxy(
            web_examples_tmpdir.virtual_snapshot,
            SOAP(web_examples_tmpdir.soap, soap_index=0),
        )

    def test_vr(self, web_examples_tmpdir):
        """
        Check that we can create a swiftgalaxy.

        We should retrieve a sample snapshot file and velociraptor catalogue.
        """
        pytest.importorskip("velociraptor")
        SWIFTGalaxy(
            web_examples_tmpdir.snapshot,
            Velociraptor(web_examples_tmpdir.velociraptor, halo_index=0),
        )

    def test_remove(self, web_examples_tmpdir):
        """Check that example data files get cleaned up on request."""
        # download all the example data (if not present)
        for example in web_examples_tmpdir.available_examples.keys():
            getattr(web_examples_tmpdir, example)

        tp = web_examples_tmpdir._demo_data_dir

        # create an additional file so that we can test empty/non-empty directory removal
        open(tp / "extra_file", "a").close()

        web_examples_tmpdir.remove()
        for absent_files in web_examples_tmpdir.available_examples.values():
            for absent_file in absent_files:
                assert not Path(absent_file).is_file()
        assert tp.is_dir()  # not yet empty
        (tp / "extra_file").unlink()
        web_examples_tmpdir.remove()
        assert not tp.is_dir()


class TestGeneratedExampleData:
    """Test the procedurally-generated example data."""

    def test_str(self, generated_examples_tmpdir):
        """Check that string representation lists available examples."""
        assert "snapshot" in str(generated_examples_tmpdir)

    def test_snapshot(self, generated_examples_tmpdir):
        """Check that we can create a swiftgalaxy with the generated snapshot helper."""
        SWIFTGalaxy(
            generated_examples_tmpdir.snapshot,
            ToyHF(snapfile=generated_examples_tmpdir.snapshot, index=0),
        )

    def test_velociraptor(self, generated_examples_tmpdir):
        """Check that we can create a swiftgalaxy using the generated velociraptor."""
        pytest.importorskip("velociraptor")
        SWIFTGalaxy(
            generated_examples_tmpdir.snapshot,
            Velociraptor(generated_examples_tmpdir.velociraptor, halo_index=0),
        )

    @pytest.mark.parametrize("group_type", ["halo", "galaxy"])
    def test_caesar(self, generated_examples_tmpdir, group_type):
        """Check that we can create a swiftgalaxy using the generated caesar."""
        pytest.importorskip("caesar")
        SWIFTGalaxy(
            generated_examples_tmpdir.snapshot,
            Caesar(
                generated_examples_tmpdir.caesar, group_type=group_type, group_index=0
            ),
        )

    def test_soap(self, generated_examples_tmpdir):
        """Check that we can create a swiftgalaxy using the generated soap."""
        pytest.importorskip("SOAP.compression")
        SWIFTGalaxy(
            generated_examples_tmpdir.virtual_snapshot,
            SOAP(generated_examples_tmpdir.soap, soap_index=0),
        )

    def test_remove(self, generated_examples_tmpdir):
        """Check that examples get cleaned up on request."""
        pytest.importorskip("SOAP.compression")
        # create all the example data (if not present)
        for example in generated_examples_tmpdir.available_examples:
            getattr(generated_examples_tmpdir, example)

        tp = generated_examples_tmpdir._demo_data_dir
        # create an additional file so that we can test empty/non-empty directory removal
        open(tp / "extra_file", "a").close()

        generated_examples_tmpdir.remove()
        absent_files = (
            tp / _toysnap_filename.name,
            tp / f"{_toyvr_filebase.name}.properties",
            tp / f"{_toyvr_filebase.name}.catalog_groups",
            tp / f"{_toyvr_filebase.name}.catalog_particles",
            tp / f"{_toyvr_filebase.name}.catalog_particles.unbound",
            tp / f"{_toyvr_filebase.name}.catalog_parttypes",
            tp / f"{_toyvr_filebase.name}.catalog_parttypes.unbound",
            tp / _toysoap_filename.name,
            tp / f"{_toysoap_membership_filebase.name}.0.hdf5",
            tp / _toysoap_virtual_snapshot_filename.name,
            tp / _toycaesar_filename.name,
        )
        for absent_file in absent_files:
            assert not Path(absent_file).is_file()
        assert tp.is_dir()  # not empty yet
        (tp / "extra_file").unlink()
        generated_examples_tmpdir.remove()
        assert not tp.is_dir()


class TestExampleNotebooks:
    """Test the example code contained in notebooks."""

    def test_generated_example_notebook(self):
        """Check that the example notebook with generated data runs without error."""
        pytest.importorskip(
            "nbmake", reason="nbmake (optional dependency) not available"
        )
        from nbmake.nb_run import NotebookRun
        from nbmake.pytest_items import NotebookFailedException

        # Currently impractical to pass values or fixtures to/from the notebook kernel.
        # Can't practically use a temporary directory for test data for the notebook.
        # So we use a somewhat ugly workaround. We set the demo data directory to a non-
        # default name to avoid colliding with user data. Then we use the `remove`
        # method to clear the directory before running the notebook to ensure we start
        # cleanly. The notebook should clean up after itself, but in case there was
        # some notebook error we tidy up afterwards.
        from swiftgalaxy.demo_data import generated_examples

        # The name set here must match the "post_cell_execute" assignment in the notebook
        # xml metadata. See https://github.com/treebeardtech/nbmake for post_cell_execute
        # explanation.
        generated_examples._demo_data_dir = Path("_pytest_notebook_demo_data")
        generated_examples.remove()

        nbr = NotebookRun(Path("examples/SWIFTGalaxy_demo.ipynb"), 300)
        result = nbr.execute()
        if result.error is not None:
            raise NotebookFailedException(result)

        generated_examples.remove()



"""Tests of string representations of objects."""

import pytest
import numpy as np
from swiftgalaxy.demo_data import web_examples, generated_examples
from swiftgalaxy.halo_catalogues import _MaskHelper


class TestStr:
    """Tests of string representations of objects and other info strings."""

    def test_coordinate_helper_str(self, sg):
        """Check that we get a sensible string representation of a coordinate helper."""
        string = str(sg.gas.spherical_coordinates)
        assert "Available coordinates:" in string
        assert "radius" in string
        assert "azimuth" in string
        assert repr(sg.gas.spherical_coordinates) == string

    def test_namedcolumn_fullname(self, sg):
        """Check _fullname of a namedcolumns matches the dataset + namedcolumns name."""
        assert (
            sg.gas.hydrogen_ionization_fractions._fullname
            == "gas.hydrogen_ionization_fractions"
        )

    def test_dataset_fullname(self, sg):
        """Check that the _fullname of a dataset matches the dataset name."""
        assert sg.gas._fullname == "gas"

    def test_sg_string(self, sg):
        """Check swiftgalaxy has a string representation (not from SWIFTDataset)."""
        string = str(sg)
        assert "SWIFTGalaxy at" in string
        assert repr(sg) == string

    @pytest.mark.parametrize("demodata", [web_examples, generated_examples])
    def test_exampledata_string(self, demodata):
        """Check that the webexample has an informative string representation."""
        string = str(demodata)
        for k in demodata.available_examples:
            assert k in string

    def test_mask_helper_repr(self):
        """Check that mask helper delegates repr to its data object."""
        expected_repr = "expected repr"

        class DummyData(object):
            def __repr__(self):
                return expected_repr

        data = DummyData()
        mask = np.s_[...]
        mh = _MaskHelper(data, mask)
        assert repr(mh) == expected_repr



"""
Configuration file for the Sphinx documentation builder.

This file only contains a selection of the most common options. For a full
list see the documentation:
https://www.sphinx-doc.org/en/master/usage/configuration.html
"""

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use pathlib to make it absolute, like shown here.
#
from pathlib import Path
import sys

sys.path.insert(0, Path("../..").absolute())
import swiftgalaxy


# -- Project information -----------------------------------------------------

project = "SWIFTGalaxy"
copyright = "2022, Kyle Oman"
author = "Kyle Oman"

# The full version, including alpha/beta/rc tags
release = swiftgalaxy.__version__


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "recommonmark",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.doctest",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx_design",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

source_suffix = [".rst", ".md"]
master_doc = "index"


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"


def setup(app):  # numpydoc ignore=GL08
    app.add_css_file("custom.css")


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

intersphinx_mapping = dict(
    swiftsimio=("https://swiftsimio.readthedocs.io/en/latest/", None),
    velociraptor=("https://velociraptor-python.readthedocs.io/en/latest/", None),
    scipy=("https://docs.scipy.org/doc/scipy/", None),
    unyt=("https://unyt.readthedocs.io/en/stable/", None),
    numpy=("https://numpy.org/doc/stable/", None),
    caesar=("https://caesar.readthedocs.io/en/latest/", None),
)



"""Test the tools for iterating over :class:`~swiftgalaxy.reader.SWIFTGalaxy`s."""

import pytest
import re
from pathlib import Path
import numpy as np
import unyt as u
from unyt.testing import assert_allclose_units
from swiftgalaxy.demo_data import (
    _toysnap_filename,
    ToyHF,
    _present_particle_types,
    _toysoap_virtual_snapshot_filename,
    _toysoap_membership_filebase,
    _toysoap_filename,
    _toyvr_filebase,
    _toycaesar_filename,
    _create_toysnap,
    _remove_toysnap,
    _create_toysoap,
    _create_toyvr,
    _create_toycaesar,
    _remove_toysoap,
    _remove_toyvr,
    _remove_toycaesar,
)
from conftest import hfs
from swiftsimio.objects import cosmo_array
from swiftgalaxy.reader import SWIFTGalaxy
from swiftgalaxy.iterator import SWIFTGalaxies
from swiftgalaxy.halo_catalogues import Standalone, SOAP, Velociraptor, Caesar


class TestSWIFTGalaxies:
    """Test the :mod:`swiftgalaxy` iteration helper class."""

    def test_eval_sparse_optimized_solution(self, toysnap):
        """
        Check that the sparse solution is chosen when optimal.

        It should match expectations for a case that we can work out by hand.
        """
        # place a single target in the centre of each cell
        # this should make sparse iteration optimal
        # at a cost of 2 cell reads
        sgs = SWIFTGalaxies(
            toysnap["toysnap_filename"],
            Standalone(
                centre=cosmo_array(
                    [[2.5, 5.0, 5.0], [7.5, 5.0, 5.0]],
                    u.Mpc,
                    comoving=True,
                    scale_factor=1.0,
                    scale_exponent=1,
                ),
                velocity_centre=cosmo_array(
                    [[0, 0, 0] * 2],
                    u.km / u.s,
                    comoving=True,
                    scale_factor=1.0,
                    scale_exponent=0,
                ),
                spatial_offsets=cosmo_array(
                    [[-1.0, 1.0], [-1.0, 1.0], [-1.0, 1.0]],
                    u.Mpc,
                    comoving=True,
                    scale_factor=1.0,
                    scale_exponent=1,
                ),
            ),
        )
        sparse_solution = sgs._sparse_optimized_solution
        assert_allclose_units(
            sparse_solution["regions"],
            cosmo_array(
                np.array(
                    [
                        [[0.01, 0.99], [0.01, 0.99], [0.01, 0.99]],
                        [[1.01, 1.99], [0.01, 0.99], [0.01, 0.99]],
                    ]
                )
                * np.array([[[5], [10], [10]]]),
                u.Mpc,
                comoving=True,
                scale_factor=1.0,
                scale_exponent=1,
            ),
            atol=0.0001 * u.Mpc,
        )
        for ss, expected in zip(
            sparse_solution["region_target_indices"], [np.array([0]), np.array([1])]
        ):
            assert np.allclose(ss, expected, atol=0)
        assert sparse_solution["cost"] == 2
        for k in sgs._solution.keys():
            if isinstance(sgs._solution[k], np.integer):
                assert sgs._solution[k] == sparse_solution[k]
            elif isinstance(sgs._solution[k], list):
                for i in range(len(sgs._solution[k])):
                    assert np.allclose(
                        sgs._solution[k][i], sparse_solution[k][i], atol=0
                    )
            else:
                assert_allclose_units(
                    sgs._solution[k], sparse_solution[k], atol=0.001 * u.Mpc
                )

    def test_eval_dense_optimized_solution(self, toysnap):
        """
        Check that the dense solution is chosen when optimal.

        It should match expectations for a case that we can work out by hand.
        """
        # Place a single target in the centre of each cell
        # and ones straddling many faces, vertices and corners
        # this should make dense iteration optimal
        # at a cost of 2-4 cell reads.
        # * Technically this solution is inefficient because
        # the cell regions are repeated, because the code is
        # not clever enough to recognize copies wrapped through
        # the periodic bounday. But having so few cells is
        # not an expected case for "real" simulations.
        # The dense solution could start to include some
        # copies if the target region is larger than ~0.5 the
        # box size in any dimension.
        sgs = SWIFTGalaxies(
            toysnap["toysnap_filename"],
            Standalone(
                centre=cosmo_array(
                    [
                        [2.5, 5.0, 5.0],
                        [5.0, 5.0, 5.0],
                        [7.5, 5.0, 5.0],
                        [5.0, 0.0, 0.0],
                        [5.0, 0.0, 9.9],
                        [5.0, 9.9, 0.0],
                        [5.0, 9.9, 9.9],
                        [9.9, 0.0, 9.9],
                        [9.9, 9.9, 0.0],
                        [9.9, 9.9, 9.9],
                    ],
                    u.Mpc,
                    comoving=True,
                    scale_factor=1.0,
                    scale_exponent=1,
                ),
                velocity_centre=cosmo_array(
                    [[0, 0, 0] * 10],
                    u.km / u.s,
                    comoving=True,
                    scale_factor=1.0,
                    scale_exponent=0,
                ),
                spatial_offsets=cosmo_array(
                    [[-1.0, 1.0], [-1.0, 1.0], [-1.0, 1.0]],
                    u.Mpc,
                    comoving=True,
                    scale_factor=1.0,
                    scale_exponent=1,
                ),
            ),
            # force this - with our 2-cell test snapshot we can't contrive for this
            # approach to be automatically determined to be optimal:
            optimize_iteration="dense",
        )
        dense_solution = sgs._dense_optimized_solution
        assert_allclose_units(
            dense_solution["regions"],
            cosmo_array(
                np.array(
                    [
                        [[0.5, 4.5], [3.0, 7.0], [3.0, 7.0]],
                        [[3.0, 11.9], [-2.0, 11.9], [-2.0, 11.9]],
                    ]
                ),
                u.Mpc,
                comoving=True,
                scale_factor=1.0,
                scale_exponent=1,
            ),
            atol=0.001 * u.Mpc,
        )
        assert len(dense_solution["region_target_indices"]) == 2
        for ds, expected in zip(
            dense_solution["region_target_indices"],
            [
                np.array([0]),
                np.array([1, 2, 3, 4, 5, 6, 7, 8, 9]),
            ],
        ):
            assert np.allclose(ds, expected, atol=0)
        # expect cost_min, cost_max = 9, 35
        # so expect cost = (9 + 35) / 2 = 22
        assert dense_solution["cost"] == 22
        for k in sgs._solution.keys():
            if isinstance(sgs._solution[k], int):
                assert sgs._solution[k] == dense_solution[k]
            elif isinstance(sgs._solution[k], list):
                for i in range(len(sgs._solution[k])):
                    assert np.allclose(
                        sgs._solution[k][i], dense_solution[k][i], atol=0
                    )
            else:
                assert_allclose_units(
                    sgs._solution[k], dense_solution[k], atol=0.001 * u.Mpc
                )

    def test_iteration_order(self, sgs):
        """Check that the iteration order is that computed for the two optimizations."""
        # force dense solution
        sgs._solution = sgs._dense_optimized_solution
        assert np.allclose(
            sgs.iteration_order,
            np.concatenate(sgs._dense_optimized_solution["region_target_indices"]),
            atol=0,
        )
        # force sparse solution
        sgs._solution = sgs._sparse_optimized_solution
        assert np.allclose(
            sgs.iteration_order,
            np.concatenate(sgs._sparse_optimized_solution["region_target_indices"]),
            atol=0,
        )

    def test_iterate(self, sgs):
        """
        Check that we iterate over the right number of SWIFTGalaxy objects.

        Each should behave like a SWIFTGalaxy created on its own for each target.
        """
        count = 0
        for sg_from_sgs in sgs:
            sg = SWIFTGalaxy(
                sgs.snapshot_filename,
                ToyHF(
                    snapfile=sgs.snapshot_filename,
                    index=sg_from_sgs.halo_catalogue.index,
                ),
            )
            for ptype in _present_particle_types.values():
                assert np.all(
                    getattr(sg_from_sgs._extra_mask, ptype).mask
                    == getattr(sg._extra_mask, ptype).mask
                )
            count += 1
        assert count == len(sgs.halo_catalogue.index)

    def test_warn_on_no_preload(self, toysnap):
        """Check that we warn users if they specify anything to pre-load (deprecated)."""
        with pytest.warns(DeprecationWarning, match="Preloading is no longer required"):
            SWIFTGalaxies(
                toysnap["toysnap_filename"],
                ToyHF(toysnap["toysnap_filename"], index=[0, 1]),
                preload={"gas.coordinates"},
            )

    def test_exception_on_repeated_targets(self, toysnap):
        """
        Check that duplicate targets are forbidden.

        Due to especially swiftsimio's masking behaviour having duplicate targets in the
        list for a SWIFTGalaxies causes all kinds of problems, so make sure we raise an
        exception if a user tries to do this.
        """
        with pytest.raises(ValueError, match="must not contain duplicates"):
            SWIFTGalaxies(
                toysnap["toysnap_filename"],
                ToyHF(toysnap["toysnap_filename"], index=[0, 0]),
            )

    def test_map(self, tmp_path_factory, hf_multi):
        """
        Check that map method returns results in the same order as the input target list.

        We're careful in this test to make sure that the iteration order is different from
        the input list order.
        """
        if isinstance(hf_multi, SOAP):
            tp = hf_multi.soap_file.parent
        elif isinstance(hf_multi, Caesar):
            tp = hf_multi.caesar_file.parent
        elif isinstance(hf_multi, Velociraptor):
            tp = Path(hf_multi.velociraptor_files["properties"]).parent
        else:
            tp = tmp_path_factory.mktemp(_toysnap_filename.parent)
            _create_toysnap(snapfile=tp / _toysnap_filename.name)
        toysnap_filename = tp / _toysnap_filename.name
        sgs = SWIFTGalaxies(
            (
                tp / _toysoap_virtual_snapshot_filename.name
                if isinstance(hf_multi, SOAP)
                else toysnap_filename
            ),
            hf_multi,
        )

        def f(sg):
            if isinstance(hf_multi, Standalone):
                return int(
                    np.argwhere(
                        np.all(
                            sg.halo_catalogue.centre == sg.halo_catalogue._centre,
                            axis=1,
                        )
                    ).squeeze()
                )
            # _index_attr has leading underscore, access through property with [1:]
            return getattr(sg.halo_catalogue, sg.halo_catalogue._index_attr[1:])

        if (np.diff(sgs.iteration_order) == 1).all():
            # if we iterate in order success is trivial, ensure that we don't:
            sgs._solution["regions"] = sgs._solution["regions"][::-1]
            sgs._solution["region_target_indices"] = sgs._solution[
                "region_target_indices"
            ][::-1]
        # double-check that success won't be trivial:
        assert not (np.diff(sgs.iteration_order) == 1).all()
        # check that map returns results ordered in input order
        if isinstance(hf_multi, Standalone):
            assert sgs.map(f) == [0, 1]
            return
        assert sgs.map(f) == getattr(
            sgs.halo_catalogue, sgs.halo_catalogue._index_attr[1:]
        )
        if isinstance(hf_multi, Standalone):
            _remove_toysnap(snapfile=toysnap_filename)

    def test_arbitrary_index_ordering(
        self, tmp_path_factory, hf_multi_forwards_and_backwards
    ):
        """
        Check that SWIFTGalaxies gives consistent results for any order of target objects.

        Especially important for velociraptor where some logic had to be added to avoid
        hdf5 complaining about an unsorted list of indices to read from file.
        """
        hf_multi, hf_multi_backwards = hf_multi_forwards_and_backwards
        if isinstance(hf_multi, SOAP):
            tp = hf_multi.soap_file.parent
        elif isinstance(hf_multi, Caesar):
            tp = hf_multi.caesar_file.parent
        elif isinstance(hf_multi, Velociraptor):
            tp = Path(hf_multi.velociraptor_files["properties"]).parent
        else:
            tp = tmp_path_factory.mktemp(_toysnap_filename.parent)
            _create_toysnap(snapfile=tp / _toysnap_filename.name)
        toysnap_filename = tp / _toysnap_filename.name

        def f(sg):
            return sg.halo_catalogue.centre

        sgs_forwards = SWIFTGalaxies(
            (
                tp / _toysoap_virtual_snapshot_filename.name
                if isinstance(hf_multi, SOAP)
                else toysnap_filename
            ),
            hf_multi,
        )
        map_forwards = sgs_forwards.map(f)
        sgs_backwards = SWIFTGalaxies(
            (
                tp / _toysoap_virtual_snapshot_filename.name
                if isinstance(hf_multi, SOAP)
                else toysnap_filename
            ),
            hf_multi_backwards,
        )
        map_backwards = sgs_backwards.map(f)
        assert np.allclose(map_forwards, map_backwards[::-1])
        if isinstance(hf_multi, Standalone):
            _remove_toysnap(snapfile=toysnap_filename)

    def test_args_kwargs_to_map(self, sgs):
        """Make sure that we can pass extra args & kwargs to a function given to map."""
        extra_arg = [("foo",), ("bar",)]
        extra_kwarg = [dict(extra_kwarg="spam"), dict(extra_kwarg="eggs")]

        def f(sg, extra_arg, extra_kwarg=None):
            return extra_arg, extra_kwarg

        result = sgs.map(f, args=extra_arg, kwargs=extra_kwarg)
        assert result == [("foo", "spam"), ("bar", "eggs")]

    def test_soap_target_order_consistency(self, toysoap_with_virtual_snapshot):
        """
        Test that SOAP doesn't sneakily reorder the galaxy catalogue.

        SOAP implicitly sorts the target mask (when getting things from the catalogue
        we get them masked in the order that they appear in the catalogue, rather
        than the order that they appear in the mask - say we ask for items [1, 0]
        as a mask, we get back those two items in order [0, 1]). This test checks
        that we get a SWIFTGalaxy from a SWIFTGalaxies that matches the one
        constructed directly when the targets given to the SWIFTGalaxies are not
        in the order that they appear in the catalogue.
        """
        soaps = [
            SOAP(
                soap_file=toysoap_with_virtual_snapshot["toysoap_filename"],
                soap_index=0,
            ),
            SOAP(
                soap_file=toysoap_with_virtual_snapshot["toysoap_filename"],
                soap_index=1,
            ),
        ]
        soap_both = SOAP(
            soap_file=toysoap_with_virtual_snapshot["toysoap_filename"],
            soap_index=[1, 0],
        )
        sgs_individual = [
            SWIFTGalaxy(
                toysoap_with_virtual_snapshot["toysoap_virtual_snapshot_filename"], soap
            )
            for soap in soaps
        ]
        sgs = SWIFTGalaxies(
            toysoap_with_virtual_snapshot["toysoap_virtual_snapshot_filename"],
            soap_both,
        )
        for i, sg in enumerate(sgs):
            sg_single = sgs_individual[sgs.iteration_order[i]]
            assert_allclose_units(
                sg_single.halo_catalogue._region_centre,
                sg.halo_catalogue._region_centre,
            )
            assert_allclose_units(
                sg_single.halo_catalogue.centre,
                sg.halo_catalogue.centre,
            )
            assert_allclose_units(
                sg_single.halo_catalogue.velocity_centre,
                sg.halo_catalogue.velocity_centre,
            )
            assert_allclose_units(
                sg_single.halo_catalogue.spherical_overdensity_200_crit.centre_of_mass,
                sg.halo_catalogue.spherical_overdensity_200_crit.centre_of_mass,
            )

    @pytest.mark.parametrize("hf_type", hfs)
    def test_halo_catalogue_with_non_list_indices(self, hf_type, toysnap_withfof):
        """
        Check that we can initialize a halo_catalogue in multi-galaxy mode.

        Here we check some ordered containers that are not a list.
        """
        toysnap_filename = toysnap_withfof["toysnap_filename"]
        tp = toysnap_filename.parent
        if hf_type == "soap":
            pytest.importorskip("SOAP.compression")
            membership_filebase = tp / _toysoap_membership_filebase.name
            toysoap_filename = tp / _toysoap_filename.name
            toysoap_virtual_snapshot_filename = (
                tp / _toysoap_virtual_snapshot_filename.name
            )
            _create_toysoap(
                filename=toysoap_filename,
                membership_filebase=membership_filebase,
                create_virtual_snapshot=True,
                create_virtual_snapshot_from=toysnap_filename,
                virtual_snapshot_filename=toysoap_virtual_snapshot_filename,
            )
        elif hf_type == "vr":
            pytest.importorskip("velociraptor")
            toyvr_filebase = tp / _toyvr_filebase.name
            _create_toyvr(filebase=toyvr_filebase)
        elif "caesar" in hf_type:
            pytest.importorskip("caesar")
            toycaesar_filename = tp / _toycaesar_filename.name
            _create_toycaesar(filename=toycaesar_filename)
        elif hf_type == "sa":
            return  # doesn't take an index list, nothing to test
        else:
            raise NotImplementedError  # a new halo_catalogue that we're not testing
        for init_indices in (
            np.array([0, 1], dtype=int),
            u.unyt_array([0, 1], u.dimensionless, dtype=int),
        ):
            if hf_type == "soap":
                sgs = SWIFTGalaxies(
                    toysoap_virtual_snapshot_filename,
                    SOAP(toysoap_filename, soap_index=init_indices),
                )
            elif hf_type == "vr":
                sgs = SWIFTGalaxies(
                    toysnap_filename,
                    Velociraptor(
                        velociraptor_filebase=toyvr_filebase, halo_index=init_indices
                    ),
                )
            elif hf_type == "caesar_galaxy":
                sgs = SWIFTGalaxies(
                    toysnap_filename,
                    Caesar(
                        toycaesar_filename,
                        group_type="galaxy",
                        group_index=init_indices,
                    ),
                )
            elif hf_type == "caesar_halo":
                sgs = SWIFTGalaxies(
                    toysnap_filename,
                    Caesar(
                        toycaesar_filename, group_type="halo", group_index=init_indices
                    ),
                )
            for sg in sgs:
                pass  # just go through the iteration
        if hf_type == "soap":
            _remove_toysoap(
                filename=toysoap_filename,
                membership_filebase=membership_filebase,
                virtual_snapshot_filename=toysoap_virtual_snapshot_filename,
            )
        elif hf_type == "vr":
            _remove_toyvr(filebase=toyvr_filebase)
        elif "caesar" in hf_type:
            _remove_toycaesar(filename=toycaesar_filename)

    def test_zero_targets(self, tmp_path_factory, hf_multi_zerotarget):
        """
        Make sure we don't crash with zero targets.

        Instead iterate over zero elements.
        """
        if isinstance(hf_multi_zerotarget, SOAP):
            tp = hf_multi_zerotarget.soap_file.parent
        elif isinstance(hf_multi_zerotarget, Caesar):
            tp = hf_multi_zerotarget.caesar_file.parent
        elif isinstance(hf_multi_zerotarget, Velociraptor):
            tp = Path(hf_multi_zerotarget.velociraptor_files["properties"]).parent
        else:
            tp = tmp_path_factory.mktemp(_toysnap_filename.parent)
            _create_toysnap(snapfile=tp / _toysnap_filename.name)
        toysnap_filename = tp / _toysnap_filename.name
        sgs = SWIFTGalaxies(
            (
                tp / _toysoap_virtual_snapshot_filename.name
                if isinstance(hf_multi_zerotarget, SOAP)
                else toysnap_filename
            ),
            hf_multi_zerotarget,
        )
        for sg in sgs:  # should not crash by iterating
            # but should not reach this:
            raise RuntimeError("Supposed to iterate over 0 elements!")

        def f(sg):
            raise RuntimeError  # we should never call this

        map_result = sgs.map(f)
        assert len(map_result) == 0

    def test_one_target(self, tmp_path_factory, hf_multi_onetarget):
        """Make sure that we don't crash with a single target. Instead iterate over it."""
        if isinstance(hf_multi_onetarget, SOAP):
            tp = hf_multi_onetarget.soap_file.parent
        elif isinstance(hf_multi_onetarget, Caesar):
            tp = hf_multi_onetarget.caesar_file.parent
        elif isinstance(hf_multi_onetarget, Velociraptor):
            tp = Path(hf_multi_onetarget.velociraptor_files["properties"]).parent
        else:
            tp = tmp_path_factory.mktemp(_toysnap_filename.parent)
            _create_toysnap(snapfile=tp / _toysnap_filename.name)
        toysnap_filename = tp / _toysnap_filename.name
        sgs = SWIFTGalaxies(
            (
                tp / _toysoap_virtual_snapshot_filename.name
                if isinstance(hf_multi_onetarget, SOAP)
                else toysnap_filename
            ),
            hf_multi_onetarget,
        )

        def f(sg):
            if isinstance(hf_multi_onetarget, Standalone):
                return int(
                    np.argwhere(
                        np.all(
                            sg.halo_catalogue.centre == sg.halo_catalogue._centre,
                            axis=1,
                        )
                    ).squeeze()
                )
            # _index_attr has leading underscore, access through property with [1:]
            return getattr(sg.halo_catalogue, sg.halo_catalogue._index_attr[1:])

        count = 0
        for sg in sgs:  # should iterate over the one target
            count += 1
            if sg.halo_catalogue._index_attr is not None:
                assert f(sg) == 1
            else:
                assert f(sg) == 0  # standalone gets position in own catalogue: 0
        assert count == 1

        map_result = sgs.map(f)
        assert len(map_result) == 1
        if sg.halo_catalogue._index_attr is not None:
            assert map_result == [1]
        else:
            assert map_result == [0]  # standalone gets position in own catalogue: 0

    def test_catalogue_not_iterable(self, toysnap):
        """Check that trying to use a non-iterable catalogue raises."""
        with pytest.raises(
            ValueError, match="halo_catalogue target list is not iterable"
        ):
            SWIFTGalaxies(
                toysnap["toysnap_filename"],
                ToyHF(snapfile=toysnap["toysnap_filename"], index=0),
            )

    def test_invalid_iteration_mode(self, toysnap):
        """Check that giving an invalid iteration mode raises."""
        with pytest.raises(ValueError, match="optimize_iteration must be one of"):
            SWIFTGalaxies(
                toysnap["toysnap_filename"],
                ToyHF(snapfile=toysnap["toysnap_filename"], index=[0, 1]),
                optimize_iteration="not_implemented",
            )

    def test_coordinate_frame_from_and_auto_recentre_invalid(self, toysnap):
        """Check inheriting coordinate frame and auto-recentering are incompatible."""
        sg = SWIFTGalaxy(
            toysnap["toysnap_filename"],
            ToyHF(snapfile=toysnap["toysnap_filename"], index=0),
        )
        sgs = SWIFTGalaxies(
            toysnap["toysnap_filename"],
            ToyHF(snapfile=toysnap["toysnap_filename"], index=[0, 1]),
            auto_recentre=True,
            coordinate_frame_from=sg,
        )
        with pytest.raises(
            ValueError, match="Cannot use coordinate_frame_from with auto_recentre"
        ):
            for sg_i in sgs:
                pass

    def test_coordinate_frame_from_in_iteration(self, toysnap):
        """Check that we can borrow a coordinate frame when iterating."""
        sg = SWIFTGalaxy(
            toysnap["toysnap_filename"],
            ToyHF(snapfile=toysnap["toysnap_filename"], index=0),
        )
        translation = cosmo_array(
            [1, 0, 0],
            u.Mpc,
            comoving=True,
            scale_factor=sg.metadata.a,
            scale_exponent=1.0,
        )
        sg.translate(translation)
        sgs = SWIFTGalaxies(
            toysnap["toysnap_filename"],
            ToyHF(snapfile=toysnap["toysnap_filename"], index=[0, 1]),
            auto_recentre=False,
            coordinate_frame_from=sg,
        )
        for sg_i in sgs:
            assert np.allclose(sg.halo_catalogue.centre - translation, sg_i.centre)

    def test_internal_units_mismatch_in_coordinate_frame_from(self, toysnap):
        """Check that incompatible internal units raises."""
        sg = SWIFTGalaxy(
            toysnap["toysnap_filename"],
            ToyHF(snapfile=toysnap["toysnap_filename"], index=0),
        )
        sg.metadata.units.length = 1 * u.kpc
        sgs = SWIFTGalaxies(
            toysnap["toysnap_filename"],
            ToyHF(snapfile=toysnap["toysnap_filename"], index=[0, 1]),
            auto_recentre=False,
            coordinate_frame_from=sg,
        )
        with pytest.raises(
            ValueError,
            match=re.escape(
                "Internal units (length and time) of coordinate_frame_from don't match."
            ),
        ):
            for sg_i in sgs:
                pass

    def test_auto_recentre_off(self, toysnap):
        """Check that we can switch of auto-recentering in iteration."""
        sgs = SWIFTGalaxies(
            toysnap["toysnap_filename"],
            ToyHF(snapfile=toysnap["toysnap_filename"], index=[0, 1]),
            auto_recentre=False,
        )
        for sg in sgs:
            assert np.allclose(sg.centre, np.zeros(3))

    def test_read_caching(self, toysnap):
        """Check that we can re-use raw data read for a region through _data_server."""
        server = SWIFTGalaxy(
            toysnap["toysnap_filename"],
            ToyHF(snapfile=toysnap["toysnap_filename"], index=0, extra_mask=None),
            auto_recentre=False,
        )
        client1 = SWIFTGalaxy(
            toysnap["toysnap_filename"],
            ToyHF(snapfile=toysnap["toysnap_filename"], index=0),
            _data_server=server,
        )
        client2 = SWIFTGalaxy(
            toysnap["toysnap_filename"],
            ToyHF(snapfile=toysnap["toysnap_filename"], index=0),
            _data_server=server,
        )
        assert client1._data_server is server
        assert client1.gas._internal_dataset._masses is None
        assert server.gas._internal_dataset._masses is None
        client1.gas.masses  # trigger read
        assert client1.gas._internal_dataset._masses is not None
        assert server.gas._internal_dataset._masses is not None
        assert (
            client1.gas._internal_dataset._masses.shape
            != server.gas._internal_dataset._masses.shape
        )
        assert not np.all(server.gas._internal_dataset._masses == 0)
        # set cached data to new values (0's) so that we can make sure that we used it
        server.gas._internal_dataset._masses[...] = 0
        assert client2._data_server is server
        assert client2.gas._internal_dataset._masses is None
        assert server.gas._internal_dataset._masses is not None
        client2.gas.masses  # trigger read: should just mask/transform cached data
        assert client2.gas._internal_dataset._masses is not None
        assert (
            client2.gas._internal_dataset._masses.shape
            != server.gas._internal_dataset._masses.shape
        )
        assert np.all(client2.gas._internal_dataset._masses == 0)



"""
Iterate over :class:`~swiftgalaxy.reader.SWIFTGalaxy` objects efficiently.

Provides the :class:`~swiftgalaxy.iterator.SWIFTGalaxies` class that enables efficient
iteration over :class:`~swiftgalaxy.reader.SWIFTGalaxy` objects for multiple objects of
interest within a single simulation snapshot.

Parallelization is not yet implemented but is prioritized for future release.
"""

import numpy as np
import unyt as u
from swiftsimio import mask, cosmo_array
from .reader import SWIFTGalaxy
from .halo_catalogues import _HaloCatalogue
from warnings import warn

from typing import Optional, Set, Any, List, Callable, Dict, Tuple, Generator, TypedDict
from swiftsimio.masks import SWIFTMask


class _IterationSolution(TypedDict):
    """Type hints for dicts containing a proposed SWIFTGalaxies iteration strategy."""

    regions: np.ndarray
    region_target_indices: list[np.ndarray]
    cost: int


class SWIFTGalaxies(object):
    """
    Facilitates efficiently iterating over many objects of interest from a simulation.

    SWIFT simulation snapshots contain particles grouped by "top-level cells" that
    cover the simulation volume. The minimum number of particles that it makes sense
    to read is therefore those contained in one such top-level cell. If one wants to
    create many :class:`~swiftgalaxy.reader.SWIFTGalaxy` objects from one simulation
    snapshot, there is a risk that the same data are read many times, such as when
    multiple target objects lie within the same top-level cell. This class provides
    a convenient way to iterate over multiple target objects while minimizing the I/O
    overhead by managing the order of iteration to group together target objects that
    occupy common top-level cells and only reading the data once.

    An important consequence to be aware of is that the iteration order is not controlled
    by the user because it must be chosen to group objects in the same top-level cell(s)
    together. The iteration order is available as the :attr:`iteration_order` attribute of
    a :class:`~swiftgalaxy.iterator.SWIFTGalaxies` object. Alternatively, output of a
    function applied to a list of target objects in the same order as the input list can
    be obtained using the :meth:`~swiftgalaxy.iterator.SWIFTGalaxies.map` method.

    There is an obvious opportunity to parallelize the iteration process by passing each
    region (potentially each containing multiple target objects) to worker processes as
    they become available, for example. This current initial version of the
    :class:`~swiftgalaxy.iterator.SWIFTGalaxies` class does not yet support parallel
    iteration, instead prioritizing the release of a working serial implementation.
    Support for parallelization will be added later as a high priority.

    Parameters
    ----------
    snapshot_filename : :obj:`str`
        Name of file containing snapshot.

    halo_catalogue : :class:`~swiftgalaxy.halo_catalogues._HaloCatalogue`
        A halo catalogue instance from :mod:`swiftgalaxy.halo_catalogues`, e.g. a
        :class:`swiftgalaxy.halo_catalogues.SOAP` instance. It should specify more
        than one target object, e.g. by setting its ``soap_index=[0, 123, 456, ...]``.

    auto_recentre : :obj:`bool` (optional), default: ``True``
        If ``True``, the coordinate system will be automatically recentred on the
        position and velocity centres defined by the ``halo_catalogue``.

    preload : set (optional), default: ``set()``
        Deprecated and ignored.

    transforms_like_coordinates : :obj:`set` (optional), default: ``set()``
        Names of fields that behave as velocities. It is assumed that these
        exist for all present particle types. When the coordinate system is
        rotated or boosted, the associated arrays will be transformed
        accordingly. The ``velocities`` dataset (or its alternative name given
        in the ``velocities_dataset_name`` parameter) is implicitly assumed to
        behave as velocities.

    transforms_like_velocities : :obj:`set` (optional), default: ``set()``
        Names of fields that behave as velocities. It is assumed that these
        exist for all present particle types. When the coordinate system is
        rotated or boosted, the associated arrays will be transformed
        accordingly. The ``velocities`` dataset (or its alternative name given
        in the ``velocities_dataset_name`` parameter) is implicitly assumed to
        behave as velocities.

    id_particle_dataset_name : :obj:`str` (optional), default: ``"particle_ids"``
        Name of the dataset containing the particle IDs, assumed to be the same
        for all present particle types.

    coordinates_dataset_name : :obj:`str` (optional), default: ``"coordinates"``
        Name of the dataset containing the particle spatial coordinates,
        assumed to be the same for all present particle types.

    velocities_dataset_name : :obj:`str` (optional), default: ``"velocities"``
        Name of the dataset containing the particle velocities, assumed to be
        the same for all present particle types.

    coordinate_frame_from : :class:`~swiftgalaxy.reader.SWIFTGalaxy` (optional), \
    default: ``None``
        Another :class:`~swiftgalaxy.reader.SWIFTGalaxy` to copy the coordinate frame
        (centre and rotation) and velocity coordinate frame (boost and rotation) from.

    optimize_iteration : :obj:`str` (optional), default: ``"auto"``
        Can be ``"auto"``, ``"dense"`` or ``"sparse"``. See docstrings of methods
        :meth:`~swiftgalaxy.iterator._eval_sparse_optimized_solution` and
        :meth:`~swiftgalaxy.iterator._eval_dense_optimized_solution` for
        explanations of optimization schemes. In most cases leave set to default
        ``"auto"`` to automatically determine optimal solution.

    Examples
    --------
    Using :class:`~swiftgalaxy.iterator.SWIFTGalaxies` is almost the same as using the
    main :class:`~swiftgalaxy.reader.SWIFTGalaxy` class, except that (i) the halo
    catalogue is initialized with multiple target objects and (ii) the
    :class:`~swiftgalaxy.iterator.SWIFTGalaxies` class provides an iteration method
    (``__iter__``), and determines its own iteration order. For example:

    ::

        from swiftgalaxy import SWIFTGalaxies, SOAP
        sgs = SWIFTGalaxies(
            "snapshot.hdf5",
            SOAP(
                "soap.hdf5",
                soap_index=[0, 123, 456],  # multiple target indices
            ),
        )
        iteration_order = sgs.iteration_order  # be aware of the order of iteration
        for sg in sgs:
            # some analysis involving the pre-loaded data fields goes here:
            sg.element_abundances.carbon
            sg.dark_matter.coordinates
            sg.stars.velocities

    Alternatively the :meth:`~swiftgalaxy.iterator.SWIFTGalaxies.map` method can be used
    to apply a function to all of the :class:`~swiftgalaxy.reader.SWIFTGalaxy`'s created
    by this class. For example:

    ::

        from swiftgalaxy import SWIFTGalaxies, SOAP
        sgs = SWIFTGalaxies(
            "snapshot.hdf5",
            SOAP(
                "soap.hdf5",
                soap_index=[0, 123, 456],  # multiple target indices
            ),
        )

        def analysis(sg):
            # this function can also have additional args & kwargs, if needed
            # it should only access the pre-loaded data fields
            sg.element_abundances.carbon
            sg.dark_matter.coordinates
            sg.stars.velocities
            return sg.element_abundances.carbon.mean()

        # map accepts arguments `args` and `kwargs`, passed through to function, if needed
        result = sgs.map(analysis)
    """

    count: int
    halo_catalogue: _HaloCatalogue

    def __init__(
        self,
        snapshot_filename: str,
        halo_catalogue: _HaloCatalogue,
        auto_recentre: bool = True,
        preload: Set[str] = set(),
        transforms_like_coordinates: Set[str] = set(),
        transforms_like_velocities: Set[str] = set(),
        id_particle_dataset_name: str = "particle_ids",
        coordinates_dataset_name: str = "coordinates",
        velocities_dataset_name: str = "velocities",
        coordinate_frame_from: Optional["SWIFTGalaxy"] = None,
        optimize_iteration: str = "auto",
    ) -> None:
        if not halo_catalogue._multi_galaxy:
            raise ValueError(
                "halo_catalogue target list is not iterable, create halo_catalogue with "
                "an iterable list of targets even if there is only one (or use "
                "SWIFTGalaxy instead of SWIFTGalaxies)."
            )
        if len(preload) > 0:
            warn(
                "Preloading is no longer required, `preload` argument will be removed in "
                "a future version.",
                DeprecationWarning,
            )
        self.halo_catalogue = halo_catalogue
        self.snapshot_filename = snapshot_filename
        self.auto_recentre = auto_recentre
        self.transforms_like_coordinates = transforms_like_coordinates
        self.transforms_like_velocities = transforms_like_velocities
        self.id_particle_dataset_name = id_particle_dataset_name
        self.coordinates_dataset_name = coordinates_dataset_name
        self.velocities_dataset_name = velocities_dataset_name
        self.coordinate_frame_from = coordinate_frame_from
        if optimize_iteration not in ("dense", "sparse", "auto"):
            raise ValueError(
                "optimize_iteration must be one of 'dense', 'sparse' or 'auto'."
            )

        if self.halo_catalogue._index_attr is not None:
            num_targets = len(
                getattr(self.halo_catalogue, self.halo_catalogue._index_attr)
            )
        else:
            num_targets = len(self.halo_catalogue._region_centre)
        # before evaluating optimized solutions:
        if num_targets == 0:
            # if we have 0 targets short-circuit
            self._solution = _IterationSolution(
                regions=np.array([]),
                region_target_indices=[np.array([])],
                cost=-1,
            )
            return
        if num_targets == 1:
            # if we have 1 target best strategy is sparse
            optimize_iteration = "sparse"

        if optimize_iteration in ("dense", "auto"):
            self._eval_dense_optimized_solution()
        if optimize_iteration in ("sparse", "auto"):
            self._eval_sparse_optimized_solution()
        if optimize_iteration == "auto":
            # typical cost of the dense solution expected to be average of the minimum
            # and maxiumum costs (one grid aligns with the cell grid), use sparse solution
            # if it costs less than this
            self._solution = (
                self._sparse_optimized_solution
                if self._sparse_optimized_solution["cost"]
                < self._dense_optimized_solution["cost"]
                else self._dense_optimized_solution
            )
        elif optimize_iteration == "dense":
            self._solution = self._dense_optimized_solution
        else:  # optimize_iteration == "sparse"
            self._solution = self._sparse_optimized_solution
        return

    @property
    def iteration_order(self) -> np.ndarray:
        """
        Property holding the order that the target objects will be iterated in.

        The iteration order is likely not the same as the order that the targets are
        provided in because this is probably not an optimal iteration order. This
        property attribute provides the optimized iteration order evaluated by
        :class:`~swiftgalaxy.iterator.SWIFTGalaxies`.

        Returns
        -------
        :class:`numpy.ndarray`
            Array of indices specifying the iteration order.
        """
        return np.concatenate(self._solution["region_target_indices"])

    def _eval_sparse_optimized_solution(self) -> None:
        """
        Evaluate an iteration scheme optimized for targets sparsely distributed in space.

        We assume that targets are sparsely distributed so that the region to be loaded
        for each one usually does not overlap with other regions. We determine the region
        to be read for each object of interest (in increments of entire top-level cells)
        and check if we can group any objects of interest together because they have
        identical regions.

        The result is stored in ``self._sparse_optimized_solution``, a dict with
        keys: ``"regions"`` containing the spatial regions that will define the spatial
        masks; ``"region_target_indices"`` containing the indices of the target objects of
        interest in each of the regions; ``"cost"`` containing the cost (in top-level cell
        read operations) of this iteration scheme.
        """
        target_centres = np.atleast_2d(self.halo_catalogue._region_centre)
        if self.halo_catalogue._user_spatial_offsets is not None:
            target_regions = self.halo_catalogue._user_spatial_offsets[np.newaxis, ...]
        else:
            aperture = self.halo_catalogue._region_aperture
            target_regions = np.vstack(
                (-np.repeat(aperture, 3), np.repeat(aperture, 3))
            ).T.reshape((-1, 3, 2))
        # SWIFTMask gives us a lightweight interface to metadata & cell metadata
        sm = mask(self.snapshot_filename, spatial_only=True)
        # get the lower cell vertex, probably at origin but not guaranteed
        cell_vertex_origin = sm.centers.min(axis=0) - sm.cell_size / 2
        # align origin with the cell grid, we allow going out the upper bounds of the
        # box since SWIFTMask handles wrapping for us (but we want a common grid
        # to easily group shared regions)
        target_centres = np.where(
            target_centres < cell_vertex_origin,
            target_centres + sm.metadata.boxsize,
            target_centres,
        )
        target_region_indices = (
            (
                (
                    target_centres[:, :, np.newaxis]
                    + target_regions
                    - cell_vertex_origin[np.newaxis, :, np.newaxis]
                )
                // sm.cell_size[np.newaxis, :, np.newaxis]
            )
            .to_value(u.dimensionless)
            .astype(int)
        )
        unique_region_indices, inv = np.unique(
            target_region_indices, axis=0, return_inverse=True
        )
        unique_regions = (
            unique_region_indices + np.array([0.01, 0.99])
        ) * sm.cell_size[np.newaxis, :, np.newaxis] + cell_vertex_origin[
            np.newaxis, :, np.newaxis
        ]
        sorter = np.argsort(inv)
        # in the following dict
        # regions are the bboxes that can be passed directly to spatial mask
        # (each one an entry along the 0th axis)
        # region_target_indices contains an array of indices into the targets array
        # for each region
        # cost is the integer number of cells that will be read during this iteration
        self._sparse_optimized_solution = _IterationSolution(
            regions=unique_regions,
            region_target_indices=np.split(
                np.arange(inv.size)[sorter],
                np.unique(inv[sorter], return_index=True)[1][1:],
            ),
            cost=np.sum(
                np.prod(
                    np.diff(unique_region_indices + np.array([0, 1]), axis=2), axis=1
                )
            ),
        )

    def _eval_dense_optimized_solution(self) -> None:
        """
        Evaluate an iteration scheme optimized for targets densely distributed in space.

        We assume that targets are densely distributed so that the region to be loaded
        for each one usually overlaps with (possibly several) other regions. We determine
        the minimum size of a region containing the largest target object and conceptually
        tile the volume with a grid with cells of this size. Each object of interest is
        assigned to the cell in which its centre lies. This groups targets into spatial
        associations. For each group of targets, the bounding box containing all of the
        targets is evaluated, resulting in a set of (probably partially overlapping)
        regions with targets assigned to each.

        The result is stored in ``self._dense_optimized_solution``, a dict with
        keys: ``"regions"`` containing the spatial regions that will define the spatial
        masks; ``"region_target_indices"`` containing the indices of the target objects of
        interest in each of the regions; ``"cost"`` containing the estimated cost (average
        of minimum and maximum cost in top-level cell read operations) of this iteration
        scheme. The actual cost depends on how each region aligns with the cell grid. For
        example if the region has a side lengths of 1.8 cells, this could touch either 2
        or 3 cells, depending on where in the grid the vertex lands.
        """
        target_centres = np.atleast_2d(self.halo_catalogue._region_centre)
        if self.halo_catalogue._user_spatial_offsets is not None:
            target_sizes = np.diff(self.halo_catalogue._user_spatial_offsets).T
        else:
            aperture = self.halo_catalogue._region_aperture
            target_sizes = np.diff(
                np.vstack((-np.repeat(aperture, 3), np.repeat(aperture, 3))).T.reshape(
                    (-1, 3, 2)
                )
            ).squeeze(2)
        # SWIFTMask gives us a lightweight interface to metadata & cell metadata
        sm = mask(self.snapshot_filename, spatial_only=True)
        # grid should be at least 1 cell in size so that we efficiently group targets
        # in the same grid location
        grid_element_dim = (
            np.max(target_sizes // sm.cell_size[np.newaxis], axis=0)
            .to_value(u.dimensionless)
            .astype(int)
            + 1
        )
        # If the grid doesn't fully cover the box need to add one more grid cell to
        # cover "too much".
        target_grid_offsets, target_grid_indices = np.modf(
            (target_centres / (grid_element_dim * sm.cell_size)).to_value(
                u.dimensionless
            )
        )
        target_grid_indices = target_grid_indices.astype(int)
        # unique_regions each correspond to a "server" SWIFTGalaxy that will read a region
        unique_grid_regions, inv = np.unique(
            target_grid_indices, axis=0, return_inverse=True
        )
        unique_regions = list()
        for unique_region_index in range(unique_grid_regions.shape[0]):
            umask = unique_region_index == inv
            masked_target_sizes = (
                target_sizes[umask]
                if self.halo_catalogue._user_spatial_offsets is None
                else target_sizes
            )
            unique_regions.append(
                np.vstack(
                    (
                        np.min(target_centres[umask] - masked_target_sizes, axis=0),
                        np.max(target_centres[umask] + masked_target_sizes, axis=0),
                    )
                ).T
            )
        unique_regions = cosmo_array(unique_regions)
        # in the following dict
        # regions are the bboxes that can be passed directly to spatial mask
        # (each one an entry along the 0th axis)
        # region_target_indices contains an array of indices into the targets array
        # for each region
        # cost_min is an integer number of cells that will be read during this iteration
        # in the best case (assuming the regions are near aligned with the cell grid)
        # cost_max is an integer number of cells that will be read during this iteration
        # in the worst case (assuming the regions are near mis-aligned with the cell grid)
        sorter = np.argsort(inv)
        cost_min = int(
            np.sum(
                np.prod(
                    np.ceil(
                        np.diff(unique_regions, axis=2).squeeze(axis=2) / sm.cell_size
                    ).to_value(u.dimensionless),
                    axis=1,
                )
            )
        )
        cost_max = int(
            np.sum(
                np.prod(
                    np.ceil(
                        np.diff(unique_regions, axis=2).squeeze(axis=2) / sm.cell_size
                    ).to_value(u.dimensionless)
                    + 1,
                    axis=1,
                )
            )
        )
        self._dense_optimized_solution = _IterationSolution(
            regions=unique_regions,
            region_target_indices=np.split(
                np.arange(inv.size)[sorter],
                np.unique(inv[sorter], return_index=True)[1][1:],
            ),
            cost=int(np.ceil(0.5 * (cost_min + cost_max))),  # estimate as the average
        )

    def _start_server(self, region_mask: SWIFTMask) -> None:
        """
        Create a "server" :class:`~swiftgalaxy.reader.SWIFTGalaxy`.

        The "server" loads all of the particles in a given region and can be repeatedly
        masked (creating copies of subsets of the particles) to efficiently provide
        :class:`~swiftgalaxy.reader.SWIFTGalaxy`'s for all objects of interest in a common
        region.

        Parameters
        ----------
        region_mask : :class:`~swiftsimio.masks.SWIFTMask`
            The spatial mask object for the server.
        """
        self._server = SWIFTGalaxy._copyinit(
            snapshot_filename=self.snapshot_filename,
            halo_catalogue=None,
            auto_recentre=False,
            transforms_like_coordinates=self.transforms_like_coordinates,
            transforms_like_velocities=self.transforms_like_velocities,
            id_particle_dataset_name=self.id_particle_dataset_name,
            coordinates_dataset_name=self.coordinates_dataset_name,
            velocities_dataset_name=self.velocities_dataset_name,
            _spatial_mask=region_mask,
            _extra_mask=None,
        )
        return

    def __iter__(self) -> Generator:
        """
        Iterate over the objects of interest defined by the halo catalogue.

        A :class:`~swiftgalaxy.iterator.SWIFTGalaxies` object is iterable, and on each
        iteration will provide one :class:`~swiftgalaxy.reader.SWIFTGalaxy` object from
        the set of objects of interest defined by the halo catalogue. The order of
        iteration is not the order that the objects are listed in the halo catalogue,
        but an optimized order determined by the
        :class:`~swiftgalaxy.iterator.SWIFTGalaxies` class. The order of iteration is
        available from :func:`~swiftgalaxy.iterator.SWIFTGalaxies.iteration_order`.
        Ordered results of applying a function to each item in the iterable set can be
        obtained using :func:`~swiftgalaxy.iterator.SWIFTGalaxies.map`.

        Yields
        ------
        :class:`~swiftgalaxy.reader.SWIFTGalaxy`
            Each subsequent :class:`~swiftgalaxy.reader.SWIFTGalaxy` object to be
            iterated over.

        See Also
        --------
        iteration_order
        map
        """
        region_mask = mask(self.snapshot_filename)
        for region, target_indices in zip(
            self._solution["regions"], self._solution["region_target_indices"]
        ):
            region_mask.constrain_spatial(region)
            self._start_server(region_mask)
            for igalaxy in target_indices:
                self.halo_catalogue._mask_multi_galaxy(igalaxy)
                server_mask = self.halo_catalogue._get_extra_mask(
                    self._server, mask_loaded=False
                )
                swift_galaxy = self._server._data_copy(
                    server_mask, _data_server=self._server
                )
                swift_galaxy.halo_catalogue = self.halo_catalogue
                if self.auto_recentre and self.coordinate_frame_from is not None:
                    raise ValueError(
                        "Cannot use coordinate_frame_from with auto_recentre=True."
                    )
                elif self.coordinate_frame_from is not None:
                    if (
                        self.coordinate_frame_from.metadata.units.length
                        != swift_galaxy.metadata.units.length
                    ) or (
                        self.coordinate_frame_from.metadata.units.time
                        != swift_galaxy.metadata.units.time
                    ):
                        raise ValueError(
                            "Internal units (length and time) of coordinate_frame_from"
                            " don't match."
                        )
                    swift_galaxy._transform(
                        self.coordinate_frame_from._coordinate_like_transform,
                        boost=False,
                    )
                    swift_galaxy._transform(
                        self.coordinate_frame_from._velocity_like_transform, boost=True
                    )
                elif self.auto_recentre:
                    swift_galaxy.recentre(self.halo_catalogue.centre)
                    swift_galaxy.recentre_velocity(self.halo_catalogue.velocity_centre)
                yield swift_galaxy
                self.halo_catalogue._unmask_multi_galaxy()

    def map(
        self,
        func: Callable,
        args: Optional[List[Tuple]] = None,
        kwargs: Optional[List[Dict]] = None,
    ) -> List[Any]:
        """
        Apply a function to each object of interest and return a list of results.

        The iteration order of :class:`~swiftgalaxy.iterator.SWIFTGalaxies` is not
        necessarily the order that the objects of interest are provided by the user
        because the class determined an efficient iteration order to minimize I/O
        operations. This method applies a provided function to each object of interest
        in an efficient order then returns the results in a list ordered in the same
        order that the objects of interest were input.

        The function to be evaluated should expect a
        :class:`~swiftgalaxy.reader.SWIFTGalaxy` (from those to be iterated over) as its
        first argument. It may accept lists of additional arguments and/or keyword
        arguments (with each element corresponding to one entry in the list of target
        objects) that can be passed to map as a :obj:`tuple` of arguments and a
        :obj:`dict` of keyword arguments.

        Currently this function only executes serially but adding a parallel execution
        option, and further support for parallelization in analysis, is a high priority.

        Parameters
        ----------
        func : callable
            The function to be evaluated.
        args : :obj:`list` (optional), default: ``None``
            List of additional arguments to the function to be evaluated (the first
            argument is always the current :class:`~swiftgalaxy.reader.SWIFTGalaxy` in the
            iteration). Each item in the list should be a :obj:`tuple` of arguments, with
            one :obj:`tuple` for each galaxy being iterated over. See examples section for
            further details.
        kwargs : :obj:`list` (optional), default: ``None``
            List of additional keyword arguments to pass to the function to be evaluated.
            Each item in the list should be a :obj:`dict` of keyword arguments, with one
            :obj:`dict` for each galaxy being iterated over. Dictionary keys are
            the names of the keyword arguments and the corresponding dictionary values are
            the values of the keyword arguments. See examples section for further details.

        Returns
        -------
        :obj:`list`
            A list containing the return value(s) of the function applied to each object
            of interest, in the same order as the objects of interest were passed to the
            halo finder interface.

        Examples
        --------
        A simple example that applies a function ``dm_median_position`` to each galaxy in
        a list of targets ``[11, 22, 33]``:

        ::

            from swiftgalaxy import SWIFTGalaxies, SOAP

            # define the function that we will apply to each SWIFTGalaxy object:
            def dm_median_position(sg):
                return np.median(sg.dark_matter.coordinates, axis=0)

            sgs = SWIFTGalaxies(
                "my_snapshot.hdf5",
                SOAP(
                    "my_soap.hdf5",
                    soap_index=[11, 22, 33],
                ),
            )
            my_result = sgs.map(dm_median_position)

        The result stored in ``my_result`` contains the result of the function for the
        galaxies at index ``11``, ``22`` and ``33``, in the same order as they are given
        in the ``soap_index`` list.

        This second example shows how to pass extra arguments and/or keyword arguments to
        the function given to ``map``:

        ::

            from swiftgalaxy import SWIFTGalaxies, SOAP

            # define the function that we will apply to each SWIFTGalaxy object:
            def dm_median_position(
                sg,  # the first argument is always a SWIFTGalaxy from the iteration
                extra_argument_1,
                extra_argument_2,
                extra_kwarg_1=None,
                extra_kwarg_2=None,
            ):
                # presumably make use of the extra arguments and/or kwargs here...
                return np.median(sg.dark_matter.coordinates, axis=0)

            sgs = SWIFTGalaxies(
                "my_snapshot.hdf5",
                SOAP("my_soap.hdf5",
                soap_index=[11, 22, 33]),
            )
            my_result = sg.map(
                dm_median_position,
                args=[
                    (my_extra_arg_1_for_galaxy_11, my_extra_arg_2_for_galaxy_11),
                    (my_extra_arg_1_for_galaxy_22, my_extra_arg_2_for_galaxy_22),
                    (my_extra_arg_1_for_galaxy_33, my_extra_arg_2_for_galaxy_33),
                ],
                kwargs=[
                    dict(
                        extra_kwarg_1=my_extra_kwarg_1_for_galaxy_11,
                        extra_kwarg_2=my_extra_kwarg_2_for_galaxy_11,
                    ),
                    dict(
                        extra_kwarg_1=my_extra_kwarg_1_for_galaxy_22,
                        extra_kwarg_2=my_extra_kwarg_2_for_galaxy_22,
                    ),
                    dict(
                        extra_kwarg_1=my_extra_kwarg_1_for_galaxy_33,
                        extra_kwarg_2=my_extra_kwarg_2_for_galaxy_33,
                    ),
                ]
            )

        Note that if you have only a single extra argument it must still be packaged as
        a tuple, for instance:

        ::

            args=[
                (my_extra_arg_for_galaxy_11, ),
                (my_extra_arg_for_galaxy_22, ),
                (my_extra_arg_for_galaxy_33, ),
            ]

        The commas inside the parentheses are not optional!
        """
        result = [None] * len(self.iteration_order)  # empty list for results
        if args is None:
            args = [tuple()] * len(self.iteration_order)
        if kwargs is None:
            kwargs = [dict()] * len(self.iteration_order)
        for sg, iteration_location in zip(self, self.iteration_order):
            result[sg.halo_catalogue._multi_galaxy_catalogue_mask] = func(
                sg, *args[iteration_location], **kwargs[iteration_location]
            )
        return result



"""Test copying instances of our classes."""

import numpy as np
from copy import copy, deepcopy
import pytest
import unyt as u
from unyt.testing import assert_allclose_units
from swiftgalaxy.masks import MaskCollection, LazyMask

abstol_m = 1e2 * u.solMass
reltol_m = 1.0e-4
abstol_nd = 1.0e-4
reltol_nd = 1.0e-4


class TestCopySWIFTGalaxy:
    """Test copying instances of :class:`~swiftgalaxy.reader.SWIFTGalaxy`."""

    def test_copy_sg(self, sg):
        """Test that dataset arrays don't get copied on shallow copy."""
        # lazy load a dataset and a named column
        sg.gas.masses
        sg.gas.hydrogen_ionization_fractions.neutral
        sg_copy = copy(sg)
        # check private attribute to not trigger lazy loading
        assert sg_copy.gas._particle_dataset._masses is None
        assert (
            sg_copy.gas.hydrogen_ionization_fractions._named_column_dataset._neutral
            is None
        )

    @pytest.mark.parametrize("derived_coords_initialized", [True, False])
    def test_deepcopy_sg(self, sg, derived_coords_initialized):
        """Test that dataset arrays get copied on deep copy."""
        # lazy load a dataset and a named column
        sg.gas.masses
        sg.gas.hydrogen_ionization_fractions.neutral
        if derived_coords_initialized:
            sg.gas.spherical_coordinates
            sg.gas.spherical_velocities
            sg.gas.cylindrical_coordinates
            sg.gas.cylindrical_velocities
        sg_copy = deepcopy(sg)
        # check private attribute to not trigger lazy loading
        assert_allclose_units(
            sg.gas.masses,
            sg_copy.gas._particle_dataset._masses,
            rtol=reltol_m,
            atol=abstol_m,
        )
        assert_allclose_units(
            sg.gas.hydrogen_ionization_fractions.neutral,
            sg_copy.gas.hydrogen_ionization_fractions._named_column_dataset._neutral,
            rtol=reltol_nd,
            atol=abstol_nd,
        )
        if derived_coords_initialized:
            assert_allclose_units(
                sg.gas.spherical_coordinates.r,
                sg_copy.gas._spherical_coordinates["_r"],
            )
            assert_allclose_units(
                sg.gas.spherical_velocities.r,
                sg_copy.gas._spherical_velocities["_v_r"],
            )
            assert_allclose_units(
                sg.gas.cylindrical_coordinates.rho,
                sg_copy.gas._cylindrical_coordinates["_rho"],
            )
            assert_allclose_units(
                sg.gas.cylindrical_velocities.rho,
                sg_copy.gas._cylindrical_velocities["_v_rho"],
            )
        else:
            assert sg_copy.gas._spherical_coordinates is None
            assert sg_copy.gas._spherical_velocities is None
            assert sg_copy.gas._cylindrical_coordinates is None
            assert sg_copy.gas._cylindrical_velocities is None


class TestCopyDataset:
    """Test copying instances of the dataset wrapper class."""

    def test_copy_dataset(self, sg):
        """Test that arrays don't get copied on shallow copy."""
        # lazy load a dataset and a named column
        sg.gas.masses
        sg.gas.hydrogen_ionization_fractions.neutral
        ds_copy = copy(sg.gas)
        # check private attribute to not trigger lazy loading
        assert ds_copy._particle_dataset._masses is None
        assert (
            ds_copy.hydrogen_ionization_fractions._named_column_dataset._neutral is None
        )

    def test_deepcopy_dataset(self, sg):
        """Test that arrays get copied on deep copy."""
        # lazy load a dataset and a named column
        sg.gas.masses
        sg.gas.hydrogen_ionization_fractions.neutral
        ds_copy = deepcopy(sg.gas)
        # check private attribute to not trigger lazy loading
        assert_allclose_units(
            sg.gas.masses,
            ds_copy._particle_dataset._masses,
            rtol=reltol_m,
            atol=abstol_m,
        )
        assert_allclose_units(
            sg.gas.hydrogen_ionization_fractions.neutral,
            ds_copy.hydrogen_ionization_fractions._named_column_dataset._neutral,
            rtol=reltol_nd,
            atol=abstol_nd,
        )


class TestCopyNamedColumns:
    """Test copying instances of the named column dataset helper class."""

    def test_copy_namedcolumn(self, sg):
        """Test that columns don't get copied on shallow copy."""
        # lazy load a named column
        sg.gas.hydrogen_ionization_fractions.neutral
        nc_copy = copy(sg.gas.hydrogen_ionization_fractions)
        # check private attribute to not trigger lazy loading
        assert nc_copy._named_column_dataset._neutral is None

    def test_deepcopy_namedcolumn(self, sg):
        """Test that columns get copied on deep copy."""
        # lazy load a named column
        sg.gas.hydrogen_ionization_fractions.neutral
        nc_copy = deepcopy(sg.gas.hydrogen_ionization_fractions)
        # check private attribute to not trigger lazy loading
        assert_allclose_units(
            sg.gas.hydrogen_ionization_fractions.neutral,
            nc_copy._named_column_dataset._neutral,
            rtol=reltol_nd,
            atol=abstol_nd,
        )


class TestCopyMaskCollection:
    """Test copying instances of the mask collection helper."""

    def test_copy_mask_collection(self):
        """Test that masks get copied."""
        mc = MaskCollection(
            gas=np.ones(100, dtype=bool),
            dark_matter=np.s_[:20],
            stars=None,
            black_holes=np.arange(3),
        )
        mc_copy = copy(mc)
        assert set(mc_copy.__dict__.keys()) == set(mc.__dict__.keys())
        for k in ("gas", "dark_matter", "stars", "black_holes"):
            comparison = getattr(mc, k) == getattr(mc_copy, k)
            if type(comparison) is bool:
                assert comparison
            else:
                assert all(comparison)

    def test_deepcopy_mask_collection(self):
        """Test that masks get copied along with values."""
        mc = MaskCollection(
            gas=np.ones(100, dtype=bool),
            dark_matter=np.s_[:20],
            stars=None,
            black_holes=np.arange(3),
            lazy_evaluated=LazyMask(mask=np.ones(100, dtype=bool)),
            lazy_unevaluated=LazyMask(mask_function=lambda: np.s_[:20]),
        )
        mc_copy = deepcopy(mc)
        assert set(mc_copy.__dict__.keys()) == set(mc.__dict__.keys())
        for k in ("gas", "dark_matter", "stars", "black_holes"):
            comparison = getattr(mc, k) == getattr(mc_copy, k)
            if type(comparison) is bool:
                assert comparison
            else:
                assert all(comparison)
        assert all(mc.lazy_evaluated.mask == mc_copy.lazy_evaluated.mask)
        assert mc.lazy_unevaluated.mask == mc_copy.lazy_unevaluated.mask


class TestCopyHaloCatalogue:
    """Test that halo catalogue classes can be copied."""

    def test_copy(self, web_hf):
        """Test that we can copy a halo catalogue object."""
        copy(web_hf)

    def test_deepcopy(self, web_hf):
        """
        Test that we can copy a halo catalogue object.

        Regression test for https://github.com/SWIFTSIM/swiftgalaxy/pull/84
        """
        deepcopy(web_hf)



"""
Tests checking that we can create objects.

If these fail something fundamental has gone wrong.
"""

from pathlib import Path
from swiftgalaxy.demo_data import _create_toyvr, _create_toycaesar, _create_toysoap
from swiftgalaxy.halo_catalogues import Caesar, Velociraptor, SOAP, Standalone
from swiftgalaxy.reader import SWIFTGalaxy
from swiftgalaxy.iterator import SWIFTGalaxies


class TestSWIFTGalaxyCreation:
    """Test that our fixtures can create the objects that we want."""

    def test_sg_creation(self, sg):
        """Make sure we can create a SWIFTGalaxy without error."""
        assert isinstance(sg, SWIFTGalaxy)

    def test_soap_creation(self, soap):
        """Make sure we can create a SOAP without error."""
        assert isinstance(soap, SOAP)

    def test_soap_recreation(self, toysoap_with_virtual_snapshot):
        """
        Make sure we can try to create a soap file set when one already exists.

        This uses the fixture. We should skip making files.
        """
        assert toysoap_with_virtual_snapshot["toysoap_filename"].is_file()
        assert Path(
            str(toysoap_with_virtual_snapshot["membership_filebase"]) + ".0.hdf5"
        ).is_file()
        assert toysoap_with_virtual_snapshot["toysnap_filename"].is_file()
        assert toysoap_with_virtual_snapshot[
            "toysoap_virtual_snapshot_filename"
        ].is_file()
        _create_toysoap(
            filename=toysoap_with_virtual_snapshot["toysoap_filename"],
            membership_filebase=toysoap_with_virtual_snapshot["membership_filebase"],
            create_membership=True,
            create_virtual_snapshot=True,
            create_virtual_snapshot_from=toysoap_with_virtual_snapshot[
                "toysnap_filename"
            ],
            virtual_snapshot_filename=toysoap_with_virtual_snapshot[
                "toysoap_virtual_snapshot_filename"
            ],
        )

    def test_vr_creation(self, vr):
        """Make sure we can create a Velociraptor without error."""
        assert isinstance(vr, Velociraptor)

    def test_vr_recreation(self, vr):
        """
        Make sure we can try to create a velociraptor file set when one already exists.

        This uses the fixture. We should skip making files.
        """
        for f in vr.velociraptor_files.values():
            assert Path(f).is_file()
        filebase = vr.velociraptor_files["properties"].split(".")[0]
        _create_toyvr(filebase=filebase)

    def test_caesar_creation(self, caesar):
        """Make sure we can create a Caesar without error."""
        assert isinstance(caesar, Caesar)

    def test_caesar_recreation(self, caesar):
        """
        Make sure we can try to create a caesar file when one already exists.

        This uses the fixture. We should skip making the file.
        """
        assert caesar.caesar_file.is_file()
        _create_toycaesar(filename=caesar.caesar_file)

    def test_sa_creation(self, sa):
        """Make sure we can create a Standalone without error."""
        assert isinstance(sa, Standalone)

    def test_sg_soap_creation(self, sg_soap):
        """Make sure we can create a SWIFTGalaxy with SOAP without error."""
        assert isinstance(sg_soap, SWIFTGalaxy)
        assert isinstance(sg_soap.halo_catalogue, SOAP)

    def test_sg_vr_creation(self, sg_vr):
        """Make sure we can create a SWIFTGalaxy with velociraptor without error."""
        assert isinstance(sg_vr, SWIFTGalaxy)
        assert isinstance(sg_vr.halo_catalogue, Velociraptor)

    def test_sg_caesar_creation(self, sg_caesar):
        """Make sure we can create a SWIFTGalaxy with Caesar without error."""
        assert isinstance(sg_caesar, SWIFTGalaxy)
        assert isinstance(sg_caesar.halo_catalogue, Caesar)

    def test_sg_sa_creation(self, sg_sa):
        """Make sure we can create a SWIFTGalaxy with Standalone without error."""
        assert isinstance(sg_sa, SWIFTGalaxy)
        assert isinstance(sg_sa.halo_catalogue, Standalone)

    def test_tab_completion(self, sg):
        """
        Check that particle dataset names and named column names are tab-completeable.

        They should be in the namespace for tab completion via the dir() method.
        """
        for prop in ("coordinates", "masses", "hydrogen_ionization_fractions"):
            # check some data fields & named column sets
            assert prop in dir(sg.gas)
        for prop in ("neutral", "ionized"):
            # check named column data fields
            assert prop in dir(sg.gas.hydrogen_ionization_fractions)
        # and check something that we inherited:
        assert "generate_empty_properties" in dir(sg.gas)
        # finally, check that we didn't lazy-load everything!
        assert sg.gas._particle_dataset._coordinates is None

    def test_internal_refs(self, sg):
        """Check that datasets and namedcolumns store a reference to their swiftgalaxy."""
        assert sg.gas._swiftgalaxy is sg
        assert sg.gas.hydrogen_ionization_fractions._swiftgalaxy is sg


class TestSWIFTGalaxiesCreation:
    """Test that our fixtures can create the objects that we want."""

    def test_sgs_creation(self, sgs):
        """Test that the fixture creates a SWIFTGalaxies."""
        assert isinstance(sgs, SWIFTGalaxies)

    def test_soap_multi_creation(self, soap_multi):
        """Test that the fixture creates a SOAP."""
        assert isinstance(soap_multi, SOAP)

    def test_vr_creation(self, vr_multi):
        """Test that the fixture creates a Velociraptor."""
        assert isinstance(vr_multi, Velociraptor)

    def test_caesar_creation(self, caesar_multi):
        """Test that the fixture creates a Caesar."""
        assert isinstance(caesar_multi, Caesar)

    def test_sa_creation(self, sa_multi):
        """Test that the fixture creates a Standalone."""
        assert isinstance(sa_multi, Standalone)

    def test_sgs_soap_creation(self, sgs_soap):
        """Test that the fixture creates a SWIFTGalaxies with SOAP."""
        assert isinstance(sgs_soap, SWIFTGalaxies)
        assert isinstance(sgs_soap.halo_catalogue, SOAP)

    def test_sgs_vr_creation(self, sgs_vr):
        """Test that the fixture creates a SWIFTGalaxies with Velociraptor."""
        assert isinstance(sgs_vr, SWIFTGalaxies)
        assert isinstance(sgs_vr.halo_catalogue, Velociraptor)

    def test_sgs_caesar_creation(self, sgs_caesar):
        """Test that the fixture creates a SWIFTGalaxies with Caesar."""
        assert isinstance(sgs_caesar, SWIFTGalaxies)
        assert isinstance(sgs_caesar.halo_catalogue, Caesar)

    def test_sgs_sa_creation(self, sgs_sa):
        """Test that the fixture creates a SWIFTGalaxies with Standalone."""
        assert isinstance(sgs_sa, SWIFTGalaxies)
        assert isinstance(sgs_sa.halo_catalogue, Standalone)


class TestDeletion:
    """Test that datasets are deleteable."""

    def test_dataset_deleter(self, sg):
        """Check that we can delete a dataset's array."""
        sg.gas.coordinates  # lazy-load some data
        assert sg.gas._internal_dataset._coordinates is not None
        del sg.gas.coordinates
        assert sg.gas._internal_dataset._coordinates is None



"""
Mask particles to select galaxies. Supplements :mod:`swiftsimio`'s spatial masking.

The :mod:`swiftsimio` masking features are optimized for spatial masking, that is
selecting regions made up of a subset of the SWIFT "top-level cells" in a simulation.
:mod:`swiftgalaxy` masking features aim to support refining these relatively coarse
spatial masks to select particles belonging to individual structures, or other
arbitrary sets of particles, and therefore needs its own masking tools. In their
long term the hope is to merge the two together, but for now the
:class:`~swiftgalaxy.masks.MaskCollection` is the recommended way to define a
selection of particles of different types for use with
:class:`~swiftgalaxy.reader.SWIFTGalaxy`.
"""

from copy import deepcopy
from typing import Optional, Union, Callable
from types import EllipsisType
from numpy.typing import ArrayLike


class LazyMask(object):
    """
    A class to hold a function to evaluate a mask until it is needed.

    This class can contain either an explicitly evaluated mask (boolean array,
    slice, etc.) or a reference to a function that returns such a mask when
    called. When the ``mask`` property is accessed, if the mask is already
    evaluated it is returned, otherwise it is evaluated and returned.

    The ``_evaluated`` attribute tracks whether the explicitly evaluated
    mask is available.

    Parameters
    ----------
    mask : slice, default: ``None``
        An object that can be used to mask an array (slice, boolean array, etc.).

    mask_function : Callable, default: ``None``
        A reference to a function that returns a mask when called.
    """

    _mask_function: Optional[Callable]
    _mask: Optional[Union[slice, EllipsisType, ArrayLike]]
    _evaluated: bool

    def __init__(
        self,
        mask: Optional[Union[slice, EllipsisType, ArrayLike]] = None,
        mask_function: Optional[Callable] = None,
    ) -> None:
        if mask_function is None and mask is None:
            self._mask = None
            self._evaluated = True
        elif mask_function is not None and mask is None:
            self._mask_function = mask_function
            self._evaluated = False
            # leave self._mask unset
        else:
            self._mask = mask
            self._mask_function = mask_function
            self._evaluated = True
        return

    def _evaluate(self) -> None:
        """Force evaluation the mask function."""
        if not self._evaluated:
            assert self._mask_function is not None  # placate mypy
            self._mask = self._mask_function()
            self._evaluated = True

    @property
    def mask(self) -> Optional[Union[slice, EllipsisType, ArrayLike]]:
        """
        Get the explicitly evaluated mask, evaluating it if necessary.

        Returns
        -------
        slice
            The explicitly evaluated mask.
        """
        if not self._evaluated:
            self._evaluate()
        return self._mask

    def __copy__(self) -> "LazyMask":
        """
        Make a copy of the :class:`~swiftgalaxy.masks.LazyMask`.

        This is without copying data (a "shallow" copy).

        Returns
        -------
        :class:`~swiftgalaxy.masks.LazyMask`
            The copy of the :class:`~swiftgalaxy.masks.LazyMask`.
        """
        if self._evaluated:
            return LazyMask(mask=self._mask, mask_function=self._mask_function)
        else:
            return LazyMask(mask_function=self._mask_function)

    def __deepcopy__(self, memo: Optional[dict] = None) -> "LazyMask":
        """
        Make a copy of the :class:`~swiftgalaxy.masks.LazyMask`.

        This copies data (a "deep" copy).

        Parameters
        ----------
        memo : :obj:`dict` (optional), default: ``None``
            For the copy operation to keep a record of already copied objects.

        Returns
        -------
        :class:`~swiftgalaxy.masks.LazyMask`
            The copy of the :class:`~swiftgalaxy.masks.LazyMask`.
        """
        if self._evaluated:
            return LazyMask(
                mask=deepcopy(self._mask), mask_function=deepcopy(self._mask_function)
            )
        else:
            return LazyMask(mask_function=deepcopy(self._mask_function))

    def __eq__(self, other: object) -> bool:
        """
        Check this mask for equality with another.

        If compared with another :class:`~swiftgalaxy.masks.LazyMask` then the comparison
        of the two explicitly evaluated masks is returned. If compared to any other
        object, comparison is attempted with the explicitly evaluated mask.

        If the mask has not been evaluated, no evaluation is triggered.

        Parameters
        ----------
        other : :obj:`object`
            The mask to compare with.

        Returns
        -------
        :obj:`bool`
            Comparison result.

        Raises
        ------
        ValueError
            If the internal mask is not already evaluated. Also raised if the compared
            object is a :class:`~swiftgalaxy.masks.LazyMask` and its mask is not
            evaluated.
        """
        if isinstance(other, LazyMask):
            if hasattr(self, "_mask") and hasattr(other, "_mask"):
                masks_equal = self._mask == other._mask
            else:
                raise ValueError(
                    "Cannot compare when one or more masks are not evaluated."
                )
        else:
            if hasattr(self, "_mask"):
                masks_equal = self._mask == other
            else:
                raise ValueError(
                    "Cannot compare when one or more masks are not evaluated."
                )
        if type(masks_equal) is not bool:
            masks_equal = all(masks_equal)
        return masks_equal

    def __ne__(self, other: object) -> bool:
        """
        Check this mask for inequality with another.

        If compared with another :class:`~swiftgalaxy.masks.LazyMask` then the comparison
        of the two explicitly evaluated masks is returned. If compared to any other
        object, comparison is attempted with the explicitly evaluated mask.

        If the mask has not been evaluated, no evaluation is triggered.

        Parameters
        ----------
        other : :obj:`object`
            The mask to compare with.

        Returns
        -------
        :obj:`bool`
            Comparison result.

        Raises
        ------
        ValueError
            If the internal mask is not already evaluated. Also raised if the compared
            object is a :class:`~swiftgalaxy.masks.LazyMask` and its mask is not
            evaluated.
        """
        return not self.__eq__(other)


class MaskCollection(object):
    """
    Barebones container for mask objects.

    Takes a set of kwargs at initialisation and assigns their values to
    attributes of the object. Attempts to access a non-existent attribute
    returns :obj:`None` instead of raising an :exc:`AttributeError`.

    This is intended to hold masks that can be applied to
    :class:`~swiftsimio.objects.cosmo_array` objects under the names of
    particle types (e.g. ``gas``, ``dark_matter``, etc.), but this is not
    checked or enforced.

    Parameters
    ----------
    **kwargs
        Any items passed as kwargs will have their values passed to
        correspondingly named attributes of this object.

    Notes
    -----

    .. note::
        The :mod:`velociraptor.swift.swift` module makes some use of a
        :obj:`namedtuple` called ``MaskCollection``. These objects are not
        valid where :mod:`swiftgalaxy` functions expect a :obj:`MaskCollection`
        because :obj:`namedtuple` objects are immutable.

    Examples
    --------
    ::

        n_dm = 123  # suppose this is number of dark matter particles
        # all these masks select all particles:
        MaskCollection(
            gas=np.s_[...],
            dark_matter=np.ones(n_dm, dtype=bool),
            stars=None
        )
    """

    def __init__(
        self,
        **kwargs: Optional[Union[slice, EllipsisType, ArrayLike, LazyMask]],
    ) -> None:
        for k, v in kwargs.items():
            if isinstance(v, LazyMask):
                setattr(self, k, v)
            elif v is None:
                pass
            else:
                setattr(self, k, LazyMask(mask=v))
        return

    def __getattr__(self, attr: str) -> None:
        """
        Return ``None`` if an attribute of the object doesn't exist.

        This function is called if an attribute is asked for and not found.
        Instead of the usual behaviour of raising a :exc:`AttributeError`,
        ``None`` is returned.

        Parameters
        ----------
        attr : :obj:`str`
            The name of the requested attribute.

        Returns
        -------
        None
            If we reach calling this function the attribute is not found and we
            return ``None``.
        """
        return None

    def __copy__(self) -> "MaskCollection":
        """
        Make a copy of the :class:`~swiftgalaxy.masks.MaskCollection`.

        This is without copying data (a "shallow" copy).

        Returns
        -------
        :class:`~swiftgalaxy.masks.MaskCollection`
            The copy of the :class:`~swiftgalaxy.masks.MaskCollection`.
        """
        return MaskCollection(**self.__dict__)

    def __deepcopy__(self, memo: Optional[dict] = None) -> "MaskCollection":
        """
        Make a copy of the :class:`~swiftgalaxy.masks.MaskCollection`.

        This copies data (a "deep" copy).

        Parameters
        ----------
        memo : :obj:`dict` (optional), default: ``None``
            For the copy operation to keep a record of already copied objects.

        Returns
        -------
        :class:`~swiftgalaxy.masks.MaskCollection`
            The copy of the :class:`~swiftgalaxy.masks.MaskCollection`.
        """
        return MaskCollection(**{k: deepcopy(v) for k, v in self.__dict__.items()})



"""Check version numbering consistency."""

import importlib.metadata
import swiftgalaxy


class TestVersion:
    """Check that version numbering is consistent."""

    def test_code_version(self):
        """Check that code version matches pyproject.toml version."""
        assert importlib.metadata.version("swiftgalaxy") == swiftgalaxy.__version__

    def test_codemeta_version(self):
        """Check that the version in codemeta.json matches pyproject.toml version."""
        with open("codemeta.json") as f:
            codemeta_content = f.read()
        assert (
            f'"version": "{importlib.metadata.version("swiftgalaxy")}",'
            in codemeta_content
        )
